#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
#
# Datasources
# Backends
from zenml.core import backends
from zenml.core.datasources.base_datasource import BaseDatasource
from zenml.core.datasources.bq_datasource import BigQueryDatasource
from zenml.core.datasources.csv_datasource import CSVDatasource
from zenml.core.datasources.image_datasource import ImageDatasource
# Pipelines
from zenml.core.pipelines.base_pipeline import BasePipeline
from zenml.core.pipelines.data_pipeline import DataPipeline
from zenml.core.pipelines.infer_pipeline import BatchInferencePipeline
from zenml.core.pipelines.training_pipeline import TrainingPipeline
from zenml.core.steps.base_step import BaseStep
from zenml.core.steps.data.base_data_step import BaseDataStep
# Steps
from zenml.core.steps.deployer import DeployerStep
from zenml.core.steps.evaluator import EvaluatorStep
from zenml.core.steps.timeseries import TimeseriesStep
# Logging
from zenml.utils.logger import init_logging

init_logging()



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base interface for BQ Data Step"""

from typing import Text, Any, Dict

import apache_beam as beam
from apache_beam.io.gcp import bigquery as beam_bigquery

from zenml.core.steps.data.base_data_step import BaseDataStep


@beam.ptransform_fn
@beam.typehints.with_input_types(beam.Pipeline)
@beam.typehints.with_output_types(beam.typehints.Dict[Text, Any])
def ReadFromBigQuery(pipeline: beam.Pipeline,
                     query_project: Text,
                     query_dataset: Text,
                     query_table: Text,
                     gcs_location: Text,
                     dest_project: Text,
                     query_limit: int = None) -> beam.pvalue.PCollection:
    query = f'SELECT * FROM `{query_project}.{query_dataset}.{query_table}`'

    if query_limit is not None:
        query += f'\nLIMIT {query_limit}'

    return (pipeline
            | 'ReadFromBigQuery' >> beam_bigquery.ReadFromBigQuery(
                project=dest_project,
                gcs_location=gcs_location,
                query=query,
                use_standard_sql=True))  # TODO: [LOW] Whats this for?


class BQDataStep(BaseDataStep):
    def __init__(self, query_project: Text, query_dataset: Text,
                 query_table: Text, gcs_location: Text,
                 dest_project: Text = None, query_limit: int = None,
                 schema: Dict = None):
        super().__init__(schema)
        self.query_project = query_project
        self.query_dataset = query_dataset
        self.query_table = query_table
        self.dest_project = dest_project
        self.query_limit = query_limit
        self.gcs_location = gcs_location

    def read_from_source(self):
        return ReadFromBigQuery(
            query_project=self.query_project,
            query_dataset=self.query_dataset,
            query_table=self.query_table,
            dest_project=self.dest_project,
            query_limit=self.query_limit,
            gcs_location=self.gcs_location,
        )



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import time

from absl import logging

from datetime import datetime
from zenml.core.pipelines.config.standard_config import MethodKeys, \
    DefaultKeys, GlobalKeys, EvaluatorKeys
import apache_beam as beam
import pytz
import tensorflow as tf

XF_SUFFIX = '_xf'


def parse_methods(input_dict, process, methods):
    """
    Args:
        input_dict:
        process:
        methods:
    """
    result = {}

    # Obtain the set of features
    f_set = set(input_dict.keys())
    f_list = list(sorted(f_set))

    for feature in f_list:
        assert isinstance(input_dict[feature], dict), \
            'Please specify a dict for every feature (empty dict if default)'

        # Check if the process for the given feature has been modified
        if process in input_dict[feature].keys():
            result[feature] = []
            for m in input_dict[feature][process]:
                MethodKeys.key_check(m)
                method_name = m[MethodKeys.METHOD]
                parameters = m[MethodKeys.PARAMETERS]
                # Check if the selected method exists and whether the
                # right parameters are given
                methods.check_name_and_params(method_name, parameters)
                result[feature].append(m)

    return result


def transformed_name(key):
    """
    Args:
        key:
    """
    return key + XF_SUFFIX


def timer(func):
    """
    Args:
        func:
    """

    def inner_func(*args, **kwargs):
        current_time = time.time()
        result = func(*args, **kwargs)
        logging.info("Function {} took {:.3f}s".format(
            func.__name__,
            time.time() - current_time
        ))
        return result

    return inner_func


def infer_schema(schema):
    """
    Args:
        schema:
    """
    schema_dict = dict()
    for feature, descriptions in schema.items():
        dtype = descriptions.dtype
        # These are ugly hacks for hard checking whether its broadly int, float
        #  bool or string.
        if dtype == float or 'float' in str(dtype):
            schema_dict[feature] = DefaultKeys.FLOAT
        elif dtype == int or 'int' in str(dtype):
            schema_dict[feature] = DefaultKeys.INTEGER
        elif dtype == bool or 'bool' in str(dtype):
            schema_dict[feature] = DefaultKeys.BOOLEAN
        elif dtype == str or 'string' in str(dtype):
            schema_dict[feature] = DefaultKeys.STRING
        else:
            raise AssertionError('Cannot infer schema dtype!')
    return schema_dict


def fill_defaults(config, default_config, schema):
    # Creating a feature spec for easier access
    """
    Args:
        config:
        default_config:
        schema:
    """
    for feature, descriptions in config.items():
        if len(descriptions) == 0:
            config[feature] = default_config[schema[feature]]
    return config


def get_function_dict(config, methods):
    """
    Args:
        config:
        methods:
    """
    result = dict()
    for feature, descriptions in config.items():
        result[feature] = []
        for d in descriptions:
            method = d[MethodKeys.METHOD]
            params = d[MethodKeys.PARAMETERS]

            result[feature].append(methods.get_method(method)(**params))

    return result


def build_eval_config(config, use_defaults=False):
    """
    Args:
        config:
        use_defaults:
    """
    import tensorflow_model_analysis as tfma
    from google.protobuf.wrappers_pb2 import BoolValue

    # SLICING SPEC
    slicing_specs = [tfma.SlicingSpec()]
    evaluator_config = config['args']

    if EvaluatorKeys.SLICES_ in evaluator_config:
        slices_config = evaluator_config[EvaluatorKeys.SLICES_]
        slicing_specs.extend([tfma.SlicingSpec(feature_keys=e)
                              for e in slices_config])

    # MODEL SPEC
    metrics_config = evaluator_config[EvaluatorKeys.METRICS_]
    metric_labels = sorted(list(set(metrics_config.keys())))

    # model_specs = [tfma.ModelSpec(
    #     signature_name='ce_eval',
    #     label_keys={'output_{}'.format(i + 1): 'label_{}_xf'.format(l)
    #                 for i, l in enumerate(metric_labels)})]

    model_specs = [tfma.ModelSpec(
        signature_name='ce_eval',
        label_keys={l: 'label_{}_xf'.format(l)
                    for i, l in enumerate(metric_labels)})]

    # METRIC SPEC
    baseline = [tfma.MetricConfig(class_name='ExampleCount')]
    metrics_specs = []
    for i, key in enumerate(metric_labels):
        metrics = baseline.copy()
        metrics.extend([tfma.MetricConfig(class_name=to_camel_case(m))
                        for m in metrics_config[key]])

        # metrics_specs.append(tfma.MetricsSpec(
        #     output_names=['output_{}'.format(i + 1)],
        #     metrics=metrics))
        metrics_specs.append(tfma.MetricsSpec(
            output_names=[key],
            metrics=metrics))

    return tfma.EvalConfig(
        model_specs=model_specs,
        slicing_specs=slicing_specs,
        metrics_specs=metrics_specs,
        options=tfma.Options(
            include_default_metrics=BoolValue(value=use_defaults)))


def to_camel_case(s):
    """
    Args:
        s:
    """
    return ''.join(list(map(lambda x: x.capitalize(), s.split('_'))))


def convert_datetime_to_secs(dt):
    """
    Args:
        dt:
    """
    delta = dt - datetime(year=1970, month=1, day=1, tzinfo=pytz.utc)
    return int(delta.total_seconds())


def df_to_example(instance) -> tf.train.Example:
    """
    Args:
        instance:
    """
    d = {col: instance[col] for col in instance.columns}

    feature = {}
    for key, value in d.items():
        if value is None:
            raise Exception('The value can not possibly be None!')
        elif value.dtype == int:
            data = value.tolist()
            feature[key] = tf.train.Feature(
                int64_list=tf.train.Int64List(value=data))
        elif value.dtype == float:
            data = value.tolist()
            feature[key] = tf.train.Feature(
                float_list=tf.train.FloatList(value=data))
        else:
            data = list(map(tf.compat.as_bytes, value.tolist()))
            feature[key] = tf.train.Feature(
                bytes_list=tf.train.BytesList(value=data))
    return tf.train.Example(features=tf.train.Features(feature=feature))


class ArrowToDataframe(beam.DoFn):
    def process(self, element, *args, **kwargs):
        """
        Args:
            element:
            *args:
            **kwargs:
        """
        return [element.to_pandas()]



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Type, Dict, List, Text

import apache_beam as beam
import numpy as np
import pandas as pd
from tensorflow_transform.tf_metadata import schema_utils
from tfx.dsl.components.base import executor_spec
from tfx.types import standard_artifacts
from tfx.types.component_spec import ComponentSpec
from tfx.types.component_spec import ExecutionParameter, ChannelParameter

from zenml.core.components.session_transform import constants as cts
# from zenml.core.components.session_transform import utils
from zenml.core.components.session_transform.component import SessionTransform
from zenml.core.components.session_transform.executor import SessionExecutor
from zenml.core.components.utils import get_function_dict, infer_schema
from zenml.core.components.utils import parse_methods, fill_defaults
from zenml.core.pipelines.config.standard_config import *

METHOD_MAPPING = {PreProcessKeys.RESAMPLING: ResamplingMethods,
                  PreProcessKeys.FILLING: FillingMethods,
                  PreProcessKeys.LABEL_TUNING: LabelTuningMethods}


class SequenceTransformSpec(ComponentSpec):
    """ComponentSpec for the PreTransform"""
    PARAMETERS = {
        cts.TIMESTAMP_COLUMN: ExecutionParameter(type=str),
        cts.CATEGORY_COLUMN: ExecutionParameter(type=str, optional=True),
        cts.FEATURES: ExecutionParameter(type=List[Text]),
        cts.LABELS: ExecutionParameter(type=List[Text]),
        cts.RESAMPLING: ExecutionParameter(type=Dict[Text, List]),
        cts.RESAMPLING + cts.DEFAULT_PREFIX: ExecutionParameter(
            type=Dict[Text, List]),
        cts.FILLING: ExecutionParameter(type=Dict[Text, List]),
        cts.FILLING + cts.DEFAULT_PREFIX: ExecutionParameter(
            type=Dict[Text, List]),
        cts.LABEL_TUNING: ExecutionParameter(type=Dict[Text, List]),
        cts.LABEL_TUNING + cts.DEFAULT_PREFIX: ExecutionParameter(
            type=Dict[Text, List]),
        cts.RESAMPLING_RATE: ExecutionParameter(type=(int, float)),
        cts.TRIP_GAP_THRESHOLD: ExecutionParameter(type=int),
        cts.SEQUENCE_LENGTH: ExecutionParameter(type=int),
        cts.SEQUENCE_SHIFT: ExecutionParameter(type=int),
    }
    INPUTS = {
        'input': ChannelParameter(type=standard_artifacts.Examples),
        'schema': ChannelParameter(type=standard_artifacts.Schema),
    }
    OUTPUTS = {
        'output': ChannelParameter(type=standard_artifacts.Examples)
    }


class SequenceTransformExecutor(SessionExecutor):
    _DEFAULT_FILENAME = 'sequence_transform'

    def get_combiner(self) -> Type[beam.CombineFn]:

        class Combiner(beam.CombineFn):
            def create_accumulator(self, *args, **kwargs):
                return pd.DataFrame()

            def add_input(self, mutable_accumulator, element, *args, **kwargs):
                return mutable_accumulator.append(element)

            def merge_accumulators(self, accumulators, *args, **kwargs):
                return pd.concat(accumulators)

            def extract_output(self, accumulator, *args, **kwargs):
                # Get the required args
                config_dict = args[0]
                s = schema_utils.schema_as_feature_spec(args[1]).feature_spec
                schema = infer_schema(s)

                # Notable columns
                timestamp_column = config_dict[cts.TIMESTAMP_COLUMN]
                category_column = config_dict[cts.CATEGORY_COLUMN]

                # Extract the values from the arrays within the cells
                session = accumulator.applymap(self._array_to_value)

                # Extract the category if there is one
                cat = None
                if category_column:
                    cat = session[category_column].unique()[0].decode('utf-8')

                # Set the timestamp as the index
                session[timestamp_column] = session[timestamp_column].apply(
                    lambda x: pd.to_datetime(x.decode('utf-8')))
                session = session.set_index(timestamp_column)

                # RESAMPLING
                resample_rate = config_dict[cts.RESAMPLING_RATE]
                resample_dict = config_dict[cts.RESAMPLING]
                resample_default_dict = config_dict[
                    cts.RESAMPLING + cts.DEFAULT_PREFIX]

                resample_dict = fill_defaults(resample_dict,
                                              resample_default_dict,
                                              schema)

                resample_function_dict = get_function_dict(resample_dict,
                                                           ResamplingMethods)

                for key, function_list in resample_function_dict.items():
                    assert len(function_list) == 1, \
                        'Please only define a single function for resampling'
                    resample_function_dict[key] = function_list[0]

                feature_list = list(resample_function_dict.keys())
                for feature in feature_list:
                    if feature not in session:
                        session[feature] = None

                resampler = session[feature_list].resample(
                    '{}L'.format(resample_rate))
                session = resampler.agg(resample_function_dict)
                session = session.dropna(
                    subset=config_dict[cts.FEATURES],
                    how='all',
                    axis=0)

                # Resolving the dtype conflicts after resampling
                output_dt_map = self._get_resample_output_dtype(resample_dict,
                                                                schema)
                session = session.astype(output_dt_map)

                # FILLING
                filling_dict = config_dict[cts.FILLING]
                filling_default_dict = config_dict[
                    cts.FILLING + cts.DEFAULT_PREFIX]

                filling_dict = fill_defaults(filling_dict,
                                             filling_default_dict,
                                             schema)

                filling_functions = get_function_dict(filling_dict,
                                                      FillingMethods)

                for key, function_list in filling_functions.items():
                    assert len(function_list) == 1, \
                        'Please only define a single function for filling'
                    filling_functions[key] = function_list[0]

                session = session.agg(filling_functions)

                # Required Keys
                if category_column:
                    session[category_column] = cat

                index_s = session.index.to_series(keep_tz=True)
                session[timestamp_column] = index_s.apply(
                    lambda x: x.strftime("%Y-%m-%d %H:%M:%S.%f %Z"))

                session[timestamp_column + '_epoch'] = \
                    index_s.apply(
                        lambda x: utils.convert_datetime_to_secs(x))

                # LABEL TUNING
                label_dict = config_dict[cts.LABEL_TUNING]
                label_defaults_dict = config_dict[
                    cts.LABEL_TUNING + cts.DEFAULT_PREFIX]

                label_dict = fill_defaults(label_dict,
                                           label_defaults_dict,
                                           schema)

                label_functions = get_function_dict(label_dict,
                                                    LabelTuningMethods)

                label_df = session[list(label_functions.keys())]
                for label, function_list in label_functions.items():
                    for f in function_list:
                        label_df.loc[:, label] = label_df.apply({label: f})

                session.update(label_df)

                session.sort_index(inplace=True)

                subseq_len = config_dict[cts.SEQUENCE_LENGTH]
                shift = config_dict[cts.SEQUENCE_SHIFT]
                return self._extract_sequences(session,
                                               subseq_len,
                                               resample_rate,
                                               shift)

            @staticmethod
            def _array_to_value(value):
                if isinstance(value, np.ndarray) and value.size == 1:
                    return value[0]
                return None

            @staticmethod
            def _get_resample_output_dtype(resampling_config, schema):
                output_dtype_map = {}
                for feature, description in resampling_config.items():
                    method = description[0][MethodKeys.METHOD]
                    params = description[0][MethodKeys.PARAMETERS]

                    output_dtype_map[
                        feature] = methods_resampling.get_output_dtype(
                        input_dtype=schema[feature],
                        method=ResamplingMethods.get_method(method),
                        params=params)

                return output_dtype_map

            @staticmethod
            def _extract_sequences(session, seq_length, window_t, shift):
                sequence_list = []
                for i in range(0, len(session.index), shift):
                    start = session.index[i]
                    end = session.index[i] + pd.to_timedelta(
                        (seq_length - 1) * window_t,
                        unit='ms')
                    ex_sequence = session.loc[start:end, :]

                    if len(ex_sequence.index) == seq_length:
                        sequence_list.append(ex_sequence)
                return sequence_list

        return Combiner


class SequenceTransform(SessionTransform):
    EXECUTOR_SPEC = executor_spec.ExecutorClassSpec(SequenceTransformExecutor)
    SPEC_CLASS = SequenceTransformSpec

    @staticmethod
    def config_parser(config):
        """
        Args:
            config:
        """
        config_dict = {k: None for k, v in
                       SequenceTransformSpec.PARAMETERS.items()
                       if v.optional}

        t_config = config[GlobalKeys.TIMESERIES_]

        # Get the categorization and index columns if they exist
        timestamp_col = t_config[TimeSeriesKeys.PROCESS_SEQUENCE_W_TIMESTAMP]
        config_dict[cts.TIMESTAMP_COLUMN] = timestamp_col
        if TimeSeriesKeys.PROCESS_SEQUENCE_W_CATEGORY_ in t_config:
            cat_col = t_config[TimeSeriesKeys.PROCESS_SEQUENCE_W_CATEGORY_]
            config_dict[cts.CATEGORY_COLUMN] = cat_col

        # Required columns for training
        feat_list = list(config[GlobalKeys.FEATURES].keys())
        label_list = list(config[GlobalKeys.LABELS].keys())

        config_dict[cts.FEATURES] = feat_list
        config_dict[cts.LABELS] = label_list

        # Feature preprocessing dictionaries
        default_config = config[GlobalKeys.PREPROCESSING]

        process_dict = {PreProcessKeys.RESAMPLING: [GlobalKeys.FEATURES,
                                                    GlobalKeys.LABELS],
                        PreProcessKeys.FILLING: [GlobalKeys.FEATURES,
                                                 GlobalKeys.LABELS],
                        PreProcessKeys.LABEL_TUNING: [GlobalKeys.LABELS]}

        for process, keys in process_dict.items():
            # For each sequential preprocessing step, extract defaults first
            config_dict.update({process: {}, process + cts.DEFAULT_PREFIX: {}})
            config_dict[process + cts.DEFAULT_PREFIX].update(parse_methods(
                default_config,
                process,
                METHOD_MAPPING[process]
            ))

            # Then extract any defined methods for that process
            for key in keys:
                assert key not in config_dict[process], \
                    'One feature can only be one of the following ' \
                    'categories: features, evaluator or labels'
                config_dict[process].update(parse_methods(
                    config[key],
                    process,
                    METHOD_MAPPING[process]))

        # Resampling rate
        resample_rate = t_config[TimeSeriesKeys.RESAMPLING_RATE_IN_SECS]
        assert isinstance(resample_rate, (int, float)), \
            '{} must be numerical'.format(
                TimeSeriesKeys.RESAMPLING_RATE_IN_SECS)
        assert resample_rate > 0, \
            '{} must be positive'.format(
                TimeSeriesKeys.RESAMPLING_RATE_IN_SECS)
        config_dict[cts.RESAMPLING_RATE] = resample_rate * 1000

        # Trip threshold
        trip_threshold = t_config[TimeSeriesKeys.TRIP_GAP_THRESHOLD_IN_SECS]
        assert isinstance(trip_threshold, int), '{} must be an integer' \
            .format(TimeSeriesKeys.TRIP_GAP_THRESHOLD_IN_SECS)
        config_dict[cts.TRIP_GAP_THRESHOLD] = trip_threshold

        # Sequence Shift
        seq_shift = t_config[TimeSeriesKeys.SEQUENCE_SHIFT]
        assert isinstance(seq_shift, int), 'sequence_shift must be an integer'
        assert seq_shift > 0, 'sequence_shift must be positive'
        config_dict[cts.SEQUENCE_SHIFT] = seq_shift

        # Sequence Length # TODO: This needs to be in the timeseries block
        seq_len = t_config[TimeSeriesKeys.SEQUENCE_LENGTH]
        assert isinstance(seq_len, int), 'sequence_length must be an integer'
        assert seq_len > 0, 'sequence_length must be positive'
        config_dict[cts.SEQUENCE_LENGTH] = seq_len

        return config_dict



# import pandas as pd
# import panel as pn
# import param
# import plotly.express as px
# import plotly.graph_objects as go
# import tensorflow_model_analysis as tfma
# # from ce_cli import constants
# # from ce_cli.evaluation import parse_metrics, get_eval_dir
# # from ce_cli.utils import api_call, api_client
#
# pn.extension('plotly')
#
#
# class Application(param.Parameterized):
#     pipeline_run_selector = param.ListSelector(default=[], objects=[
#     ])  # 1
#     hyperparameter_selector = param.ListSelector(default=[],
#                                                  objects=[])  # 2
#
#     slicing_metric_selector = param.ObjectSelector(default='',
#                                                    objects=[''])  # 3
#     performance_metric_selector = param.ObjectSelector(objects=[])  # 4
#
#     def __init__(self, **params):
#         super(Application, self).__init__(**params)
#         # apis
#         p_api = ce_api.PipelinesApi(api_client(info))
#         ws_api = ce_api.WorkspacesApi(api_client(info))
#
#         user = info[constants.ACTIVE_USER]
#         active_workspace = info[user][constants.ACTIVE_WORKSPACE]
#
#         # lists
#         result_list = []
#         hparam_list = []
#
#         # get all pipelines in this workspace
#         all_pipelines = api_call(
#             ws_api.get_workspaces_pipelines_api_v1_workspaces_workspace_id_pipelines_get,
#             active_workspace)
#
#         # get a dataframe of all results + all hyperparameter combinations
#         for p in all_pipelines:
#             if p.pipeline_runs is not None and len(p.pipeline_runs) > 0:
#                 for run in p.pipeline_runs:
#                     # This is slowing the comparison down but
#                     # necessary to update the status of each run
#                     r = api_call(
#                         p_api.get_pipeline_run_api_v1_pipelines_pipeline_id_runs_pipeline_run_id_get,
#                         p.id,
#                         run.id)
#                     if r.status == 'Succeeded':
#                         eval_path = get_eval_dir(p.id, r.id, info)
#                         evaluation = tfma.load_eval_result(eval_path)
#                         for s, m in evaluation.slicing_metrics:
#                             result_list.append(dict([(
#                                 'pipeline_name', '{}:{}'.format(p.name,
#                                                                 r.id)),
#                                 ('slice_name',
#                                  s[0][0] if s
#                                  else ''),
#                                 ('slice_value',
#                                  s[0][1] if s
#                                  else ''),
#                                 ('metrics',
#                                  m[''][''])]))
#
#                         h_dict = api_call(
#                             p_api.get_hyperparameters_pipeline_api_v1_pipelines_pipeline_id_runs_pipeline_run_id_hyperparameters_get,
#                             p.id,
#                             r.id)
#                         h_dict['pipeline_name'] = '{}:{}'.format(
#                             p.name, r.id)
#                         hparam_list.append(h_dict)
#
#         self.results = pd.DataFrame([parse_metrics(r) for r in
#                                      result_list])
#         self.hparam_info = pd.DataFrame(hparam_list)
#
#         # set params
#         self.param.pipeline_run_selector.objects = self.results[
#             'pipeline_name'].unique()
#
#     def generate_results_with_additional_params(self, df,
#                                                 track_list,
#                                                 pipeline_run_selector):
#         print('Generating results...')
#         tracking_dict = dict()
#         for key in track_list:
#             component, parameter = key.split(':')
#             if component in tracking_dict:
#                 tracking_dict[component].append(parameter)
#             else:
#                 tracking_dict[component] = [parameter]
#
#         context_list = list(self.results[self.results[
#             'pipeline_name'].isin(pipeline_run_selector)][
#                                 'context_id'].unique())
#         for c in context_list:
#             for e in api_get_list_of_executions(int(c), info)[0]:
#                 component_id = e['properties']['component_id'][
#                     'stringValue']
#                 if component_id in tracking_dict:
#                     parameter_list = tracking_dict[component_id]
#                     for p in parameter_list:
#                         param_name = ':'.join([component_id, p])
#                         param_value = e['properties'][p]['stringValue']
#                         df.loc[df['context_id'] == c, param_name] =
#                         float(param_value)
#
#         return df
#
#     @param.depends('pipeline_run_selector', watch=True)
#     def _updated_context(self):
#         df = self.results[self.results['pipeline_name'].isin(
#             self.pipeline_run_selector)]
#         df = df.dropna(axis=1, how='all')
#
#         slicing_metric_list = sorted(list(df['slice_name'].unique()))
#
#         performance_metric_set = {c for c in df.columns if
#                                   c.startswith('metric_')}
#         performance_metric_list = [None] + sorted(list(
#             performance_metric_set))
#
#         self.param['slicing_metric_selector'].objects = slicing_metric_list
#         self.param[
#             'performance_metric_selector'].objects = performance_metric_list
#
#         # get params
#         parameter_list = self.hparam_info[self.hparam_info[
#             'pipeline_name'].isin(self.pipeline_run_selector)].columns
#         parameter_list = [x for x in parameter_list if x !=
#                           'pipeline_name']
#         self.param['hyperparameter_selector'].objects = parameter_list
#
#         self.slicing_metric_selector = ''
#         self.performance_metric_selector = None
#
#     @param.depends('slicing_metric_selector',
#                    'performance_metric_selector', watch=True)
#     def performance_graph(self):
#         if self.performance_metric_selector:
#             df = self.results[(self.results['pipeline_name'].isin(
#                 self.pipeline_run_selector)) &
#                               (self.results['slice_name'] ==
#                                self.slicing_metric_selector)]
#             fig = px.scatter(df,
#                              x='pipeline_name',
#                              y=self.performance_metric_selector,
#                              color='slice_value',
#                              width=1100,
#                              title='Pipeline Comparison')
#
#             fig = fig.update_traces(mode='lines+markers')
#
#         else:
#             fig = px.scatter(pd.DataFrame(),
#                              marginal_y='rug',
#                              width=1100,
#                              title='Pipeline Comparison')
#
#         return fig
#
#     @param.depends('performance_metric_selector',
#                    'hyperparameter_selector', watch=True)
#     def parameter_graph(self):
#         if self.performance_metric_selector and len(
#                 self.hyperparameter_selector) > 0:
#             df = self.results[(self.results['pipeline_name'].isin(
#                 self.pipeline_run_selector)) &
#                               (self.results['slice_name'] == '')]
#
#             # merge
#             extra_df = pd.merge(self.hparam_info, df,
#                                 on='pipeline_name', how='left')
#
#             dimensions = ['pipeline_name'] + self.hyperparameter_selector + \
#                          [self.performance_metric_selector]
#
#             new_dims = []
#             for d in dimensions:
#                 try:
#                     new_dims.append({'label': d, 'values':
#                         pd.to_numeric(extra_df[d])})
#                 except:
#                     u = sorted(list(extra_df[d].apply(lambda x: str(
#                         x)).unique()))
#                     mapping = {v: i for i, v in enumerate(u)}
#                     new_dims.append({'label': d,
#                                      'tickvals': [mapping[x] for x
#                                                   in u],
#                                      'ticktext': u,
#                                      'values': extra_df[d].apply(
#                                          lambda x: str(x)).map(mapping)})
#
#             final_col = pd.to_numeric(extra_df[
#                                           self.performance_metric_selector])
#             fig = go.Figure(data=go.Parcoords(line=dict(
#                 color=final_col,
#                 colorscale=
#                 'inferno',
#                 showscale=True,
#                 cmin=min(
#                     final_col),
#                 cmax=max(
#                     final_col)),
#                 dimensions=new_dims))
#         else:
#             fig = px.scatter(pd.DataFrame(),
#                              marginal_y='rug',
#                              width=1100,
#                              title='Hyperparameter Comparison')
#
#         return fig
#
#
# def generate_interface():
#     app = Application()
#     handlers = pn.Param(app.param)
#
#     # Analysis Page
#     analysis_page = pn.GridSpec(height=850, width=1850, max_height
#     =850, max_width=1850)
#     analysis_page[0:8, 0:2] = handlers[1]
#     analysis_page[0:10, 8:10] = handlers[2]
#     analysis_page[8:9, 0:2] = handlers[3]
#     analysis_page[9:10, 0:2] = handlers[4]
#     analysis_page[0:5, 2:8] = app.performance_graph
#     analysis_page[5:10, 2:8] = app.parameter_graph
#
#     interface = pn.Tabs(
#         ('Analysis Page', analysis_page),
#     )
#     return interface
#
#
# platform = generate_interface()
# platform.servable()



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import tensorflow_transform as tft

from zenml.core.components.session_transform.methods import \
    methods_resampling, \
    methods_label_tuning, methods_filling
from zenml.core.components.transform.methods import methods_nonseq_filling, \
    methods_transform
from zenml.core.pipelines.config.config_utils import ConfigKeys, \
    MethodDescriptions


class GlobalKeys(ConfigKeys):
    SPLIT = 'split'
    FEATURES = 'features'
    LABELS = 'labels'
    TRAINER = 'trainer'
    EVALUATOR = 'evaluator'
    PREPROCESSING = 'preprocessing'
    TIMESERIES_ = 'timeseries'
    PCA_ = 'pca'
    CUSTOM_CODE_ = 'custom_code'

    # following keys are MANUALLY added to the config during pipeline creation
    VERSION = 'version'
    BQ_ARGS_ = 'bq_args'


class DataSourceKeys(ConfigKeys):
    DATA_TYPE = 'type'
    DATA_SOURCE = 'source'
    ARGS = 'args'


class BQArgsKeys(ConfigKeys):
    PROJECT = 'project'
    DATASET = 'dataset'
    TABLE = 'table'


class GCSKeys(ConfigKeys):
    PATH = 'path'
    SERVICE_ACCOUNT = 'service_account'


class TimeSeriesKeys:
    RESAMPLING_RATE_IN_SECS = 'resampling_rate_in_secs'
    TRIP_GAP_THRESHOLD_IN_SECS = 'trip_gap_threshold_in_secs'

    PROCESS_SEQUENCE_W_TIMESTAMP = 'process_sequence_w_timestamp'
    PROCESS_SEQUENCE_W_CATEGORY_ = 'process_sequence_w_category'

    SEQUENCE_SHIFT = 'sequence_shift'
    SEQUENCE_LENGTH = 'sequence_length'


class SplitKeys(ConfigKeys):
    RATIO_ = "ratio"
    CATEGORIZE_BY_ = 'categorize'
    INDEX_BY_ = 'index'
    WHERE_ = 'where'


class EvaluatorKeys(ConfigKeys):
    SLICES_ = 'slices'
    METRICS_ = 'metrics'


class PCAKeys(ConfigKeys):
    NUM_DIMENSIONS = 'num_dimensions'


class TrainerKeys(ConfigKeys):
    FN = 'fn'
    PARAMS = 'params'


class DefaultKeys(ConfigKeys):
    STRING = 'string'
    INTEGER = 'integer'
    BOOLEAN = 'boolean'
    FLOAT = 'float'


class PreProcessKeys(ConfigKeys):
    RESAMPLING = 'resampling'
    FILLING = 'filling'
    TRANSFORM = 'transform'
    LABEL_TUNING = 'label_tuning'


class CustomCodeKeys(ConfigKeys):
    TRANSFORM_ = 'transform'
    MODEL_ = 'model'


class CustomCodeMetadataKeys(ConfigKeys):
    NAME = 'name'
    UDF_PATH = 'udf_path'
    STORAGE_PATH = 'storage_path'


class MethodKeys(ConfigKeys):
    METHOD = 'method'
    PARAMETERS = 'parameters'


#########################

class ResamplingMethods(MethodDescriptions):
    MODES = {'mean': (methods_resampling.resample_mean, []),
             'threshold': (methods_resampling.resample_thresholding,
                           ['cond', 'c_value', 'threshold', 'set_value']),
             'mode': (methods_resampling.resample_mode, []),
             'median': (methods_resampling.resample_median, [])}


class FillingMethods(MethodDescriptions):
    MODES = {'forward': (methods_filling.forward_f, []),
             'backwards': (methods_filling.backwards_f, []),
             'min': (methods_filling.min_f, []),
             'max': (methods_filling.max_f, []),
             'mean': (methods_filling.mean_f, []),
             'custom': (methods_filling.custom_f, ['custom_value'])}


class LabelTuningMethods(MethodDescriptions):
    MODES = {'leadup': (methods_label_tuning.tune_leadup,
                        ['event_value', 'duration']),
             'followup': (methods_label_tuning.tune_followup,
                          ['event_value', 'duration']),
             'shift': (methods_label_tuning.tune_shift,
                       ['shift_steps', 'fill_value']),
             'map': (methods_label_tuning.tune_map, ['mapper']),
             'no_tuning': (methods_label_tuning.no_tune, [])}


class TransformMethods(MethodDescriptions):
    MODES = {'scale_by_min_max': (tft.scale_by_min_max, ['min', 'max']),
             'scale_to_0_1': (tft.scale_to_0_1, []),
             'scale_to_z_score': (tft.scale_to_z_score, []),
             'tfidf': (tft.tfidf, ['vocab_size']),
             'compute_and_apply_vocabulary': (
                 tft.compute_and_apply_vocabulary, []),
             'ngrams': (tft.ngrams, ['ngram_range', 'separator']),
             'hash_strings': (tft.hash_strings, ['hash_buckets']),
             'pca': (tft.pca, ['output_dim', 'dtype']),
             'bucketize': (tft.bucketize, ['num_buckets']),
             'no_transform': (lambda x: x, []),
             'image_reshape': (methods_transform.image_reshape, ['height',
                                                                 'width',
                                                                 'num_channels'
                                                                 ]),
             'load_binary_image': (methods_transform.load_binary_image,
                                   []),
             'one_hot_encode': (methods_transform.one_hot_encode, ['depth'])}


class NonSeqFillingMethods(MethodDescriptions):
    MODES = {'min': (methods_nonseq_filling.min_f, []),
             'max': (methods_nonseq_filling.max_f, []),
             'mean': (methods_nonseq_filling.mean_f, []),
             'custom': (methods_nonseq_filling.custom_f, ['custom_value'])}


# class TrainerMethods(MethodDescriptions):
#     MODES = {'feedforward': (methods_trainer.feedforward, [])}



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os

from zenml.utils.enums import GDPComponent

APP_NAME = 'zenml'
CONFIG_VERSION = '1'

COMPARISON_NOTEBOOK = 'comparison_notebook.ipynb'
EVALUATION_NOTEBOOK = 'evaluation_notebook.ipynb'

ACTIVE_USER = '_active_user'
ACTIVE_CONFIG = 'active_config'
ACTIVE_WORKSPACE = 'active_workspace'
ACTIVE_DATASOURCE_COMMIT = 'active_datasource_commit'
ACTIVE_ORCHESTRATOR = 'active_orchestrator'
ACTIVE_PROVIDER = 'active_provider'
TOKEN = 'token'

# env variables that modules consume from
ZENML_PIPELINE_CONFIG = 'ZENML_PIPELINE_CONFIG'
CE_CLOUD_JOB_PREFIX = 'CE_CLOUD_JOB_PREFIX'
ENV_KUBEFLOW_GCR = 'KUBEFLOW_GCR'
ENV_TRAINER_GCR = 'TRAINER_GCR'
ENV_ZENML_DEBUG = 'ZENML_DEBUG'
ENV_ZENML_LOGGING_VERBOSITY = 'ZENML_LOGGING_VERBOSITY'

# images kubeflow sources from
GDP_IMAGE = os.getenv(ENV_KUBEFLOW_GCR)
TRAINER_IMAGE = os.getenv(ENV_TRAINER_GCR)

# TODO: [HIGH] Switch to default False before release
IS_DEBUG_ENV = os.getenv(ENV_ZENML_DEBUG, True)
if IS_DEBUG_ENV:
    ZENML_LOGGING_VERBOSITY = os.getenv(ENV_ZENML_LOGGING_VERBOSITY, 4)
else:
    ZENML_LOGGING_VERBOSITY = os.getenv(ENV_ZENML_LOGGING_VERBOSITY, 0)


# Pipeline related parameters
PREPROCESSING_FN = 'gdp.components.transform.transform_module.preprocessing_fn'
TRAINER_FN = 'gdp.components.trainer.trainer_module.run_fn'


# IMPORTANT: This should be ONLY Alphabets AND LESS than 30 characters long!
# NO other characters allowed. This is because the dataflow jobs are named
# after them to match these names while calculating cost and billing!


def get_component_mapping():
    from tfx.components.evaluator.component import Evaluator
    from tfx.components.pusher.component import Pusher
    from tfx.components.schema_gen.component import SchemaGen
    from tfx.components.statistics_gen.component import StatisticsGen

    from zenml.core.components.session_transform.sequence_transform import \
        SequenceTransform
    from zenml.core.components.split_gen.component import SplitGen
    from zenml.core.components.trainer.component import Trainer
    from zenml.core.components.transform.component import Transform
    from zenml.core.components.data_gen.component import DataGen

    return {
        GDPComponent.SplitGen.name: SplitGen.__name__,
        GDPComponent.SplitStatistics.name: StatisticsGen.__name__,
        GDPComponent.SplitSchema.name: SchemaGen.__name__,
        GDPComponent.SequenceTransform.name: SequenceTransform.__name__,
        GDPComponent.SequenceStatistics.name: StatisticsGen.__name__,
        GDPComponent.SequenceSchema.name: SchemaGen.__name__,
        GDPComponent.PreTransform.name: 'PreTransform',  # TODO: TBI
        GDPComponent.PreTransformStatistics.name: StatisticsGen.__name__,
        GDPComponent.PreTransformSchema.name: SchemaGen.__name__,
        GDPComponent.Transform.name: Transform.__name__,
        GDPComponent.Trainer.name: Trainer.__name__,
        GDPComponent.Evaluator.name: Evaluator.__name__,
        GDPComponent.Deployer.name: Pusher.__name__,
        GDPComponent.DataGen.name: DataGen.__name__,
    }



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import pandas as pd


def tune_leadup(event_value, duration):
    """
    Args:
        event_value:
        duration:
    """
    def apply(x):
        timings = x.loc[x == event_value].index
        result = pd.Series(0, index=x.index)
        for t in timings:
            start = t - pd.to_timedelta(duration, unit='s')
            result.loc[start:t] = 1
        return result

    return apply


def tune_followup(event_value, duration):
    """
    Args:
        event_value:
        duration:
    """
    def apply(x):
        timings = x.loc[x == event_value].index
        result = pd.Series(0, index=x.index)
        for t in timings:
            end = t + pd.to_timedelta(duration, unit='s')
            result.loc[t:end] = 1
        return result

    return apply


def tune_shift(shift_steps, fill_value):
    """
    Args:
        shift_steps:
        fill_value:
    """
    def apply(x):
        return x.shift(shift_steps, fill_value=fill_value)

    return apply


def tune_map(mapper):
    """Replaces values of column specified through the mapper dict

    mapper: dict where key is the value to be replaced and value is the value
    to used for replacement

    Args:
        mapper:
    """

    def apply(x):
        for orig_val, new_val in mapper.items():
            x = x.replace(orig_val, str(new_val))
        return x

    return apply


def no_tune():
    def apply(x):
        return x

    return apply



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import apache_beam as beam

from zenml.core.components.data_gen.types.interface import TypeInterface


class TextConverter(TypeInterface):
    @staticmethod
    def convert(pipeline, output_dict, exec_properties):
        """
        Args:
            pipeline:
            output_dict:
            exec_properties:
        """
        result = (pipeline
                  | beam.Map(lambda x: x)
                  )
        return result



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

def create_new_cell(contents):
    """

    Args:
        contents:

    Returns:

    """
    import sys
    from IPython.core.getipython import get_ipython

    if 'ipykernel' not in sys.modules:
        raise EnvironmentError('The magic functions are only usable in a '
                               'Jupyter notebook.')

    shell = get_ipython()

    payload = dict(
        source='set_next_input',
        text=contents,
        replace=False,
    )
    shell.payload_manager.write_payload(payload, single=False)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from zenml.utils.source_utils import load_source_path_class


def preprocessing_fn(inputs, custom_config):
    c = load_source_path_class(custom_config['source'])

    if 'features' in custom_config['args']:
        features = custom_config['args']['features']
    else:
        features = None

    if 'labels' in custom_config['args']:
        labels = custom_config['args']['labels']
    else:
        labels = None

    if 'overwrite' in custom_config['args']:
        overwrite = custom_config['args']['overwrite']
    else:
        overwrite = None

    # TODO: decide what to do with the additional params
    return c(features=features,
             labels=labels,
             overwrite=overwrite).get_preprocessing_fn()(inputs)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Dict, Text, Any, List

from tfx.components.bulk_inferrer.component import BulkInferrer
from tfx.components.common_nodes.importer_node import ImporterNode
from tfx.types import standard_artifacts

import zenml.core.pipelines.config.standard_config as exp_keys
from zenml.core.backends.orchestrator.local.orchestrator_local_backend import \
    OrchestratorLocalBackend
from zenml.core.backends.processing.processing_local_backend import \
    ProcessingLocalBackend
from zenml.core.components.session_transform.sequence_transform import \
    SequenceTransform
from zenml.core.components.split_gen.component import SplitGen
from zenml.core.pipelines.base_pipeline import BasePipeline
from zenml.utils.enums import GDPComponent


class BatchInferencePipeline(BasePipeline):
    """BatchInferencePipeline definition to run batch inference pipelines.

    A BatchInferencePipeline is used to run an inference based on a
    TrainingPipeline.
    """
    PIPELINE_TYPE = 'infer'

    def __init__(self, name: Text, *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        raise NotImplementedError('Coming soon!')

    def get_tfx_component_list(self, spec: Dict[Text, Any]) -> List:
        """
        Args:
            spec:
        """
        component_list = list()

        # Create the single split
        # splits = SplitGen(config=train_config,
        #                   instance_name=GDPComponent.SplitGen.name)
        # component_list.append(splits)
        # datapoints = splits.outputs.examples

        # TODO [MEDIUM] Handle timeseries when added
        # if exp_keys.GlobalKeys.TIMESERIES_ in train_config:
        #     schema = ImporterNode(instance_name='Schema',
        #                           source_uri=spec['schema_uri'],
        #                           artifact_type=standard_artifacts.Schema)

            # sequence_transform = SequenceTransform(
            #     examples=datapoints,
            #     schema=schema,
            #     config=train_config,
            #     instance_name=GDPComponent.SequenceTransform.name)
            # datapoints = sequence_transform.outputs.output
            # component_list.extend([schema, sequence_transform])

        # Infer the results
        # model = ImporterNode(instance_name='Trainer',
        #                      source_uri=spec['model_uri'],
        #                      artifact_type=standard_artifacts.Model)
        #
        # bulk_inferrer = BulkInferrer(examples=datapoints,
        #                              model=model.outputs.result,
        #                              instance_name=GDPComponent.Inferrer.name)
        #
        # component_list.extend([model, bulk_inferrer])
        #
        # return component_list

    def get_default_backends(self) -> Dict:
        """Gets list of default backends for this pipeline."""
        # For base class, orchestration is always necessary
        return {
            OrchestratorLocalBackend.BACKEND_KEY: OrchestratorLocalBackend(),
            ProcessingLocalBackend.BACKEND_KEY: ProcessingLocalBackend()
        }



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import abc
import logging
import os
from typing import Any, Dict, List, Text, Type

import apache_beam as beam
import dateutil
import pytz
from absl import logging
from apache_beam.transforms.window import Sessions, GlobalWindows
from apache_beam.transforms.window import TimestampedValue
from apache_beam.utils.timestamp import Timestamp
from tensorflow_data_validation.coders import tf_example_decoder
from tfx import types
from tfx.components.base import base_executor
from tfx.types import artifact_utils
from tfx.utils import io_utils

from zenml.core.components.session_transform import constants
# from zenml.core.components.session_transform import utils


class SessionExecutor(base_executor.BaseExecutor):
    """Baseline Executor for the SessionTransform component

    It includes a beam pipeline, which windows into sessions with a given gap
    for each split and applies a CombineFn to the sessions.

    In order to use it, follow the instruction given in the component.py
    """
    _DEFAULT_FILENAME = 'session_transform_data'

    def Do(self,
           input_dict: Dict[Text, List[types.Artifact]],
           output_dict: Dict[Text, List[types.Artifact]],
           exec_properties: Dict[Text, Any]) -> None:
        # Start the log
        """
        Args:
            input_dict:
            output_dict:
            exec_properties:
        """
        self._log_startup(input_dict, output_dict, exec_properties)

        # Configuration
        trip_gap_threshold = exec_properties[constants.TRIP_GAP_THRESHOLD]

        # Names of relevant columns
        category_column = exec_properties[constants.CATEGORY_COLUMN]
        timestamp_column = exec_properties[constants.TIMESTAMP_COLUMN]

        # Get each split
        input_artifact = artifact_utils.get_single_instance(
            input_dict['input'])
        split_names = artifact_utils.decode_split_names(
            input_artifact.split_names)

        # Get the schema
        schema_path = io_utils.get_only_uri_in_dir(
            artifact_utils.get_single_uri(input_dict['schema']))
        schema = io_utils.SchemaReader().read(schema_path)

        logging.warning(self._beam_pipeline_args)
        with self._make_beam_pipeline() as p:
            # Iterate Through Each Split
            for split in split_names:
                # Logging
                logging.info('SequenceTransform on the {} split'.format(split))

                input_uri = io_utils.all_files_pattern(
                    artifact_utils.get_split_uri(input_dict['input'], split))

                output_uri = artifact_utils.get_split_uri(
                    output_dict['output'],
                    split)

                output_path = os.path.join(output_uri, self._DEFAULT_FILENAME)

                # Read and decode the data
                data = \
                    (p
                     | 'Read_' + split >> beam.io.ReadFromTFRecord(
                                file_pattern=input_uri)
                     | 'Decode_' + split >>
                     tf_example_decoder.DecodeTFExample()
                     | 'ArrowToDataFrame_' + split >> beam.ParDo(
                                utils.ArrowToDataframe()))

                # Window into sessions
                s_data = \
                    (data
                     | 'AddCategory_' + split >> beam.ParDo(self.AddKey(),
                                                            category_column)
                     | 'AddTimestamp_' + split >> beam.ParDo(
                                self.AddTimestamp(),
                                timestamp_column)
                     | 'Sessions_' + split >> beam.WindowInto(
                                Sessions(trip_gap_threshold))
                     )

                # Combine and transform
                p_data = \
                    (s_data
                     | 'Combine_' + split >> beam.CombinePerKey(
                                self.get_combiner()(),
                                exec_properties,
                                schema)
                     )

                # Write the results
                _ = \
                    (p_data
                     | 'Global_' + split >> beam.WindowInto(GlobalWindows())
                     | 'RemoveKey_' + split >> beam.ParDo(self.RemoveKey())
                     | 'ConvertToExample_' + split >> beam.Map(
                                utils.df_to_example)
                     | 'SerializeDeterministically_' + split >> beam.Map(
                                lambda x: x.SerializeToString(
                                    deterministic=True))
                     | 'WriteSplit_' + split >> beam.io.WriteToTFRecord(
                                output_path, file_name_suffix='.gz'))

    class AddKey(beam.DoFn):
        def process(self, element, category_column):
            """Add the category as a key to the rows

            Args:
                element:
                category_column:
            """
            if category_column:
                element = element.set_index(
                    element[category_column].apply(
                        lambda x: x[0].decode('utf-8')),
                    drop=False)
            else:
                element[constants.DEFAULT_CAT] = constants.DEFAULT_CAT
                element = element.set_index(constants.DEFAULT_CAT)
            return element.iterrows()

    class RemoveKey(beam.DoFn):
        def process(self, element):
            """Remove the key from the output of the CombineFn and return the
            resulting datapoints

            Args:
                element:
            """
            return element[1]

    class AddTimestamp(beam.DoFn):
        @staticmethod
        def extract_t(element, timestamp_column):
            """
            Args:
                element:
                timestamp_column:
            """
            timestamp = dateutil.parser.parse(element[timestamp_column][0])
            timestamp_utc = timestamp.astimezone(pytz.utc)
            return Timestamp.from_utc_datetime(timestamp_utc)

        def process(self, element, timestamp_column):
            """Parse the timestamp and add it to the datapoints

            Args:
                element:
                timestamp_column:
            """
            unix_timestamp = self.extract_t(element[1], timestamp_column)
            yield TimestampedValue(element, unix_timestamp)

    @abc.abstractmethod
    def get_combiner(self) -> Type[beam.CombineFn]:
        """Abstract method which needs to be implemented to with the transform
        logic on the session level
        """
        pass



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os

import absl
import apache_beam as beam
from absl import logging
from tfx.orchestration import pipeline
from tfx.orchestration.beam.beam_dag_runner import BeamDagRunner, \
    _ComponentAsDoFn
from tfx.orchestration.config import config_utils
from tfx.utils import telemetry_utils


class ZenMLBeamlDagRunner(BeamDagRunner):
    """
    This is the same as the super class from tfx:
    tfx.orchestration.beam.beam_dag_runner.BeamDagRunner with the exception
    being that the pipeline_run is not overridden. Full credit to Google LLC
    for the code.
    """
    def run(self, tfx_pipeline: pipeline.Pipeline) -> None:
        if 'TFX_JSON_EXPORT_PIPELINE_ARGS_PATH' in os.environ:
            return

        with telemetry_utils.scoped_labels(
                {telemetry_utils.LABEL_TFX_RUNNER: 'beam'}):
            with beam.Pipeline(argv=self._beam_orchestrator_args) as p:
                root = p | 'CreateRoot' >> beam.Create([None])

                signal_map = {}
                for component in tfx_pipeline.components:
                    component_id = component.id

                    signals_to_wait = []
                    if component.upstream_nodes:
                        for upstream_node in component.upstream_nodes:
                            assert upstream_node in signal_map, (
                                'Components is not in '
                                'topological order')
                            signals_to_wait.append(
                                signal_map[upstream_node])
                    absl.logging.info('Component %s depends on %s.',
                                      component_id,
                                      [s.producer.full_label for s in
                                       signals_to_wait])

                    (component_launcher_class,
                     component_config) = \
                        config_utils.find_component_launch_info(
                            self._config, component)

                    signal_map[component] = (
                            root
                            | 'Run[%s]' % component_id >> beam.ParDo(
                        _ComponentAsDoFn(component,
                                         component_launcher_class,
                                         component_config, tfx_pipeline),
                        *[beam.pvalue.AsIter(s) for s in signals_to_wait]))
                    absl.logging.info('Component %s is scheduled.',
                                      component_id)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Dict, Text, Any

import apache_beam as beam

from zenml.core.components.data_gen.constants import DestinationKeys
from zenml.core.components.data_gen.types.interface import TypeInterface
from zenml.core.components.data_gen.utils import DtypeInferrer
from zenml.core.components.data_gen.utils import append_tf_example
from zenml.core.components.data_gen.utils import init_bq_table


@beam.ptransform_fn
@beam.typehints.with_input_types(beam.typehints.Dict[Text, Any])
@beam.typehints.with_output_types(beam.pvalue.PDone)
def WriteToBQ(datapoints: Dict[Text, Any],
              schema: Dict[Text, Any],
              destination: Dict[Text, Any]) -> beam.pvalue.PDone:
    # Obtain the schema
    """
    Args:
        datapoints:
        schema:
        destination:
    """
    if schema:
        schema_dict = schema
        assert all(v in ['STRING', 'INTEGER', 'FLOAT', 'TIMESTAMP',
                         'BOOLEAN', 'BYTES']
                   for _, v in schema.items())
        init_bq_table(
            destination[DestinationKeys.PROJECT],
            destination[DestinationKeys.DATASET],
            destination[DestinationKeys.TABLE],
            schema)
    else:

        schema = (datapoints
                  | 'InferSchema' >> beam.CombineGlobally(
                    DtypeInferrer(),
                    project=destination[DestinationKeys.PROJECT],
                    dataset=destination[DestinationKeys.DATASET],
                    table=destination[DestinationKeys.TABLE]))
        schema_dict = beam.pvalue.AsSingleton(schema)

    return (datapoints
            | 'AddTFExample' >> beam.Map(append_tf_example, schema_dict)
            | 'WriteToBQ' >> beam.io.WriteToBigQuery(
                table=destination[DestinationKeys.TABLE],
                dataset=destination[DestinationKeys.DATASET],
                project=destination[DestinationKeys.PROJECT],
                write_disposition='WRITE_EMPTY')
            )


class TabularConverter(TypeInterface):
    def convert(self):
        return WriteToBQ



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Definition of the base Orchestrator Backend"""

from typing import Dict, Text, Any

from tfx.orchestration import pipeline

from zenml.core.backends.base_backend import BaseBackend
from zenml.core.backends.orchestrator.local.zenml_local_orchestrator import \
    ZenMLLocalDagRunner
from zenml.core.standards.standard_config import EnvironmentKeys


class OrchestratorLocalBackend(BaseBackend):
    """Base class for all ZenML backends.

    Every ZenML pipeline runs in backends
    """
    BACKEND_TYPE = 'local'
    BACKEND_KEY = 'orchestrator'

    @staticmethod
    def get_tfx_pipeline(config: Dict[Text, Any]) -> pipeline.Pipeline:
        """
        Converts ZenML config dict to TFX pipeline.

        Args:
            config: a ZenML config dict

        Returns: a TFX pipeline object.
        """
        from zenml.core.pipelines.base_pipeline import BasePipeline
        zenml_pipeline = BasePipeline.from_config(config)
        component_list = zenml_pipeline.get_tfx_component_list(config)
        spec = zenml_pipeline.get_pipeline_spec(config)

        tfx_pipeline = pipeline.Pipeline(
            components=component_list,
            beam_pipeline_args=spec['execution_args'],
            metadata_connection_config=spec['metadata_connection_config'],
            pipeline_name=spec['repo_id'],
            pipeline_root=spec['pipeline_root'],
            log_root=spec['pipeline_log_root'],
            enable_cache=spec[EnvironmentKeys.ENABLE_CACHE])

        # Ensure that the run_id is ZenML pipeline_name
        tfx_pipeline.pipeline_info.run_id = zenml_pipeline.pipeline_name
        return tfx_pipeline

    def run(self, config: Dict[Text, Any]):
        """
        This run function essentially calls an underlying TFX orchestrator run.
        However it is meant as a higher level abstraction with some
        opinionated decisions taken.

        Args:
            config: a ZenML config dict
        """
        tfx_pipeline = self.get_tfx_pipeline(config)
        ZenMLLocalDagRunner().run(tfx_pipeline)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Dict

import tensorflow as tf

from zenml.core.components.utils import infer_schema, transformed_name
from zenml.core.components.utils import parse_methods
from zenml.core.standards.standard_config import MethodKeys
from zenml.core.standards.standard_methods import NonSeqFillingMethods
from zenml.core.standards.standard_methods import TransformMethods
from zenml.core.steps.preprocesser.base_preprocesser import \
    BasePreprocesserStep

DEFAULT_DICT = {
    'boolean': {
        'filling': [{'method': 'max', 'parameters': {}}],
        'resampling': [{'method': 'threshold',
                        'parameters': {'c_value': 0,
                                       'cond': 'greater',
                                       'set_value': 1,
                                       'threshold': 0}}],
        'transform': [{'method': 'no_transform', 'parameters': {}}]},
    'float': {
        'filling': [{'method': 'max', 'parameters': {}}],
        'resampling': [{'method': 'mean', 'parameters': {}}],
        'transform': [{'method': 'scale_to_z_score', 'parameters': {}}]},
    'integer': {
        'filling': [{'method': 'max', 'parameters': {}}],
        'resampling': [{'method': 'mean', 'parameters': {}}],
        'transform': [{'method': 'scale_to_z_score', 'parameters': {}}]},
    'string': {
        'filling': [{'method': 'custom', 'parameters': {'custom_value': ''}}],
        'resampling': [{'method': 'mode', 'parameters': {}}],
        'transform': [{'method': 'compute_and_apply_vocabulary',
                       'parameters': {}}]
    }}


class PreprocesserStep(BasePreprocesserStep):
    def __init__(self, features=None, labels=None, overwrite=None,
                 **unused_kwargs):
        """
        Base level preprocesser step for generic preprocessing.
        
        Args:
            features:
            labels:
            overwrite:
            **unused_kwargs:
        """
        super(PreprocesserStep, self).__init__()

        # List of features and labels
        self.features = features or []
        self.labels = labels or []

        self.t_dict = parse_methods(overwrite or {},
                                    'transform',
                                    TransformMethods)

        self.t_d_dict = parse_methods(DEFAULT_DICT,
                                      'transform',
                                      TransformMethods)

        self.f_dict = parse_methods(overwrite or {},
                                    'filling',
                                    NonSeqFillingMethods)

        self.f_d_dict = parse_methods(DEFAULT_DICT,
                                      'filling',
                                      NonSeqFillingMethods)

    def preprocessing_fn(self, inputs: Dict):
        schema = infer_schema(inputs)

        output = {}
        for key, value in inputs.items():
            if key in self.features or key in self.labels:

                if key not in self.f_dict:
                    self.f_dict[key] = self.f_d_dict[schema[key]]
                value = self.apply_filling(value, self.f_dict[key])

                if key not in self.t_dict:
                    self.t_dict[key] = self.t_d_dict[schema[key]]
                result = self.apply_transform(key, value, self.t_dict[key])
                result = tf.cast(result, dtype=tf.float32)

                if key in self.features:
                    output[transformed_name(key)] = result

                if key in self.labels:
                    output[transformed_name('label_' + key)] = result

            output[key] = value

        return output

    @staticmethod
    def apply_transform(key, data, transform_list):
        for transform in transform_list:
            method_name = transform[MethodKeys.METHOD]
            method_params = transform[MethodKeys.PARAMETERS]

            if method_name == 'compute_and_apply_vocabulary':
                method_params.update({'vocab_filename': key})

            data = TransformMethods.get_method(method_name)(data,
                                                            **method_params)
        return data

    @staticmethod
    def apply_filling(data, filling_list):
        method_name = filling_list[0][MethodKeys.METHOD]
        method_params = filling_list[0][MethodKeys.PARAMETERS]
        return NonSeqFillingMethods.get_method(method_name)(**method_params)(
            data)

    def get_preprocessing_fn(self):
        return self.preprocessing_fn



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import re
from typing import Text, Dict, Any

import apache_beam as beam
from google.cloud import storage
from google.oauth2 import service_account as sa_service
from tfx.types.component_spec import ExecutionParameter

from zenml.core.components.data_gen.sources.interface import SourceInterface


@beam.ptransform_fn
@beam.typehints.with_input_types(beam.Pipeline)
@beam.typehints.with_output_types(beam.typehints.Dict[Text, Any])
class BlobToBytes(beam.DoFn):
    def __init__(self, service_account):
        """This is the only way to not do crazy client instantiation cycles.

        Args:
            service_account:
        """
        self._sa = service_account

    def parse_gcs_path(self, gcs_path, object_optional=False):
        """Return the bucket and object names of the given gs:// path.

        Args:
            gcs_path:
            object_optional:
        """
        match = re.match('^gs://([^/]+)/(.*)$', gcs_path)
        if match is None or (match.group(2) == '' and not object_optional):
            raise ValueError(
                'GCS path must be in the form gs://<bucket>/<object>.')
        return match.group(1), match.group(2)

    def start_bundle(self):
        self.credentials = sa_service.Credentials.from_service_account_info(
            self._sa)
        self.gcs_client = storage.Client(credentials=self.credentials)

    def process(self, element):
        """Parse the blob and convert to TFExample

        Args:
            element:
        """
        bucket_name, blob_name = self.parse_gcs_path(element)
        bucket = self.gcs_client.bucket(bucket_name)
        blob = bucket.blob(blob_name)
        blob_as_bstring = blob.download_as_string()
        return [blob_as_bstring]


class GCSInjection(SourceInterface):
    def get_config_spec(self):
        return {'input_path': ExecutionParameter(type=Text),
                'service_account': ExecutionParameter(type=Dict[Text, Any])}

    def config_parser(self, config):
        """
        Args:
            config:
        """
        return config

    def beam_transform(self, pipeline, exec_properties):
        """
        Args:
            pipeline:
            exec_properties:
        """
        input_path = exec_properties['input_path']
        service_account = exec_properties['service_account']

        result = (pipeline
                  | beam.io.ReadFromText(input_path)
                  | beam.ParDo(BlobToBytes(service_account))
                  )
        return result



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the categorical domain split."""

from typing import Text, Dict, List, Any

from zenml.core.steps.split.utils import get_categorical_value
from zenml.core.steps.split import constants
from zenml.core.steps.split.base_split_step import BaseSplitStep


def lint_split_map(split_map: Dict[Text, List[Text]]):
    """Small utility to lint the split_map"""
    if constants.TRAIN not in split_map.keys():
        raise AssertionError(f'You have to define some values for '
                             f'the {constants.TRAIN} split.')
    if len(split_map) <= 1:
        raise AssertionError('Please specify more than 1 split name in the '
                             'split_map!')


def CategoricalPartitionFn(element: Any,
                           num_partitions: int,
                           categorical_column: Text,
                           split_map: Dict[Text, List[Text]]) -> int:
    """
    Function for a categorical split on data to be used in a beam.Partition.
    Args:
        element: Data point, given as a tf.train.Example.
        num_partitions: Number of splits, unused here.
        categorical_column: Name of the categorical column in the data on which
        to perform the split.
        split_map: Dict {split_name: [category_list]} mapping the categorical
        values in categorical_column to their respective splits.

    Returns: An integer n, where 0 <= n <= num_partitions - 1.

    """
    category_value = get_categorical_value(element, cat_col=categorical_column)

    # The following code produces a dict: { split_name: unique_integer }
    # However 'train' always maps to 0
    enumerated_splits = {constants.TRAIN: 0}
    enumerated_splits.update(
        {name: i for i, name in enumerate(split_map.keys())
         if name != constants.TRAIN}
    )

    for split_name, category_value_list in split_map.items():
        # if the value is in the list, then just return
        if category_value in category_value_list:
            # return the index of that split
            return enumerated_splits[split_name]

    # This is the default behavior for category_values that dont belong to 
    #  any split in the split_map. We assign it to index '1'
    return 1


class CategoricalDomainSplitStep(BaseSplitStep):

    def __init__(
            self,
            categorical_column: Text,
            split_map: Dict[Text, List[Text]],
            statistics=None,
            schema=None,
    ):
        """

        Args:
            statistics: Parsed statistics artifact from a preceding
            StatisticsGen.
            schema: Parsed schema artifact from a preceding SchemaGen.
            categorical_column: Name of the categorical column in the data on
            which to split.
            split_map: a dict { split_name: [category_values] }.
        """
        self.categorical_column = categorical_column

        lint_split_map(split_map)
        self.split_map = split_map

        super().__init__(statistics, schema)

    def partition_fn(self):
        return CategoricalPartitionFn, {
            'split_map': self.split_map,
            'categorical_column': self.categorical_column
        }

    def get_split_names(self) -> List[Text]:
        return list(self.split_map.keys())



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from zenml.core.components.data_gen import constants
from zenml.core.components.data_gen.types.images import ImageConverter
from zenml.core.components.data_gen.types.segmented_images import CVImageConverter
from zenml.core.components.data_gen.types.tabular import TabularConverter
from zenml.core.components.data_gen.types.text import TextConverter


class TypeFactory:

    def __init__(self):
        self.types = {}

    def get_types(self):
        return self.types

    def get_single_type(self, key):
        """
        Args:
            key:
        """
        return self.types[key]

    def register_type(self, key, type_):
        """
        Args:
            key:
            type_:
        """
        self.types[key] = type_


# Register the injections into the factory
type_factory = TypeFactory()
type_factory.register_type(constants.TYPE_TEXT, TextConverter)
type_factory.register_type(constants.TYPE_TABULAR, TabularConverter)
type_factory.register_type(constants.TYPE_IMAGE, ImageConverter)
type_factory.register_type(constants.TYPE_IMAGE_SEG, CVImageConverter)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Definition of the DataFlow Processing Backend"""

from typing import Text

from zenml.core.backends.processing.processing_local_backend import \
    ProcessingLocalBackend


class ProcessingDataFlowBackend(ProcessingLocalBackend):
    """Base class for all Processing ZenML backends.

    Every ZenML pipeline runs in backends.
    """
    BACKEND_TYPE = 'dataflow'

    def __init__(
            self, worker_machine_type: Text = 'n1-standard-4',
            num_workers: int = 4,
            max_num_workers: int = 10,
            disk_size_gb: int = 100,
            autoscaling_algorithm: Text = 'THROUGHPUT_BASED',
            **kwargs):
        self.worker_machine_type = worker_machine_type
        self.num_workers = num_workers
        self.max_num_workers = max_num_workers
        self.disk_size_gb = disk_size_gb
        self.autoscaling_algorithm = autoscaling_algorithm
        super().__init__(**kwargs)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
from typing import Text, Dict, Any

import apache_beam as beam
from tfx.utils import io_utils
from tfx_bsl.coders import csv_decoder

from zenml.core.components.data_gen.constants import LocalCSVArgs
from zenml.core.components.data_gen.sources.interface import SourceInterface


@beam.ptransform_fn
@beam.typehints.with_input_types(beam.Pipeline)
@beam.typehints.with_output_types(beam.typehints.Dict[Text, Any])
def ReadFilesFromDisk(pipeline: beam.Pipeline,
                      source_args: Dict[Text, Any]) -> beam.pvalue.PCollection:
    """
    Args:
        pipeline (beam.Pipeline):
        source_args:
    """
    base_path = source_args['base_path']

    wildcard_qualifier = "*"

    # we are ingesting all the files from the base path by
    # supplying the wildcard (hopefully?)
    file_pattern = os.path.join(base_path, wildcard_qualifier)

    import tensorflow as tf
    csv_files = tf.io.gfile.glob(file_pattern)
    if not csv_files:
        raise RuntimeError(
            'Split pattern {} does not match any files.'.format(file_pattern))

    column_names = io_utils.load_csv_column_names(csv_files[0])

    parsed_csv_lines = (
            pipeline
            | 'ReadFromText' >> beam.io.ReadFromText(file_pattern=file_pattern,
                                                     skip_header_lines=1)
            | 'ParseCSVLine' >> beam.ParDo(csv_decoder.ParseCSVLine(
        delimiter=','))
            | 'ExtractParsedCSVLines' >> beam.Map(lambda x:
                                                  dict(zip(column_names, x))))

    return parsed_csv_lines


class LocalCSVSource(SourceInterface):
    ARG_KEYS = LocalCSVArgs

    def beam_transform(self):
        return ReadFilesFromDisk



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Pipeline to create data sources"""

from typing import Dict, Text, Any, List

from tfx.components.schema_gen.component import SchemaGen
from tfx.components.statistics_gen.component import StatisticsGen

from zenml.core.backends.orchestrator.local.orchestrator_local_backend import \
    OrchestratorLocalBackend
from zenml.core.backends.processing.processing_local_backend import \
    ProcessingLocalBackend
from zenml.core.components.data_gen.component import DataGen
from zenml.core.components.data_gen.constants import SpecParamKeys
from zenml.core.pipelines.base_pipeline import BasePipeline
from zenml.core.pipelines.standards.standard_training_pipeline import \
    GlobalKeys
from zenml.core.standards import standard_config as keys
from zenml.utils.enums import GDPComponent
from zenml.utils.post_training.post_training_utils import \
    get_statistics_artifact, \
    get_schema_artifact, view_statistics, view_schema


class DataPipeline(BasePipeline):
    """DataPipeline definition to create datasources.

    A DataPipeline is used to create datasources in ZenML. Each data pipeline
    creates a snapshot of the datasource in time. All datasources are consumed
    by different ZenML pipelines like the TrainingPipeline.
    """
    PIPELINE_TYPE = 'data'

    def get_tfx_component_list(self, config: Dict[Text, Any]) -> List:
        """
        Creates a data tfx pipeline.

        Args:
            config: a ZenML config as a dict
        """
        data_config = config[GlobalKeys.STEPS][keys.DataSteps.DATA]
        data_gen = DataGen(
            source=data_config[SpecParamKeys.SOURCE],
            source_args=data_config[SpecParamKeys.SOURCE_ARGS]).with_id(
            GDPComponent.DataGen.name
        )

        datapoints = data_gen.outputs.examples

        statistics_split = StatisticsGen(
            examples=datapoints).with_id(GDPComponent.SplitStatistics.name)

        schema_split = SchemaGen(
            statistics=statistics_split.outputs.output,
            infer_feature_shape=True).with_id(GDPComponent.SplitSchema.name)

        return [data_gen, statistics_split, schema_split]

    def view_statistics(self, magic: bool = False):
        """
        View statistics for data pipeline in HTML.

        Args:
            magic (bool): Creates HTML page if False, else
            creates a notebook cell.
        """
        uri = get_statistics_artifact(
            self.pipeline_name, GDPComponent.DataStatistics.name)
        view_statistics(uri, magic)

    def view_schema(self):
        """View schema of data flowing in pipeline."""
        uri = get_schema_artifact(
            self.pipeline_name, GDPComponent.DataSchema.name)
        view_schema(uri)

    def get_default_backends(self) -> Dict:
        """Gets list of default backends for this pipeline."""
        # For base class, orchestration is always necessary
        return {
            OrchestratorLocalBackend.BACKEND_KEY: OrchestratorLocalBackend(),
            ProcessingLocalBackend.BACKEND_KEY: ProcessingLocalBackend()
        }

    def steps_completed(self) -> bool:
        mandatory_steps = [keys.DataSteps.DATA]
        for step_name in mandatory_steps:
            if step_name not in self.steps_dict.keys():
                raise AssertionError(
                    f'Mandatory step {step_name} not added.')
        return True



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
import re
from typing import Text, Dict, Any
from uuid import uuid4
import oauth2client.service_account

import apache_beam as beam
from apache_beam.io.gcp import gcsio
from google.cloud import storage
from apache_beam.internal.http_client import get_new_http
from apache_beam.io.gcp.internal.clients.storage import StorageV1
from google.oauth2 import service_account as sa_service
from tfx.types.component_spec import ExecutionParameter
from tfx_bsl.coders import csv_decoder

from zenml.core.components.data_gen.sources.interface import SourceInterface


# @beam.ptransform_fn
# @beam.typehints.with_input_types(beam.Pipeline)
# @beam.typehints.with_output_types(beam.typehints.Dict[Text, Any])
class ProduceGCSBlobs(beam.DoFn):
    def __init__(self, service_account, *args, **kwargs):
        """This is the only way to not do crazy client instantiation cycles.

        Args:
            service_account:
            *args:
            **kwargs:
        """
        self._sa = service_account

    def parse_gcs_path(self, gcs_path, object_optional=False):
        """Return the bucket and object names of the given gs:// path.

        Args:
            gcs_path:
            object_optional:
        """
        match = re.match('^gs://([^/]+)/(.*)$', gcs_path)
        if match is None or (match.group(2) == '' and not object_optional):
            raise ValueError(
                'GCS path must be in the form gs://<bucket>/<object>.')
        return match.group(1), match.group(2)

    def start_bundle(self):
        self.credentials = sa_service.Credentials.from_service_account_info(
            self._sa)
        sa = self._sa
        scope = "https://www.googleapis.com/auth/devstorage.read_only"
        self.credentials = oauth2client.service_account.\
            ServiceAccountCredentials.from_json_keyfile_dict(keyfile_dict=sa,
                                                             scopes=scope)
        self.gcs_client = StorageV1(credentials=self.credentials,
                                http=get_new_http(),
                                response_encoding="utf8")
        self.gcs_conn = gcsio.GcsIO(storage_client=self.gcs_client)

    def process(self, element, *args, **kwargs):
        """Parse the blob and download as bytestring

        Args:
            element:
            *args:
            **kwargs:
        """
        # bucket_name, blob_name = self.parse_gcs_path(element)
        # bucket = self.gcs_client.bucket(bucket_name)
        # blob = bucket.blob(blob_name)
        # blob_as_bstring = blob.download_as_string()
        # yield element, blob_as_bstring
        yield self.gcs_conn.open(element).read().decode("utf-8")


class CSVColumnInferrer(beam.DoFn):
    def __init__(self, service_account, max_header_size=16384, *unused_args,
                 **unused_kwargs):
        """Infer csv columns from the blob. Works for headers up to 16
        kilobytes.

        Args:
            service_account:
            max_header_size:
            *unused_args:
            **unused_kwargs:
        """
        super().__init__(*unused_args, **unused_kwargs)
        self._sa = service_account
        self.max_header_size = max_header_size

    def parse_gcs_path(self, gcs_path, object_optional=False):
        """Return the bucket and object names of the given gs:// path.

        Args:
            gcs_path:
            object_optional:
        """
        match = re.match('^gs://([^/]+)/(.*)$', gcs_path)
        if match is None or (match.group(2) == '' and not object_optional):
            raise ValueError(
                'GCS path must be in the form gs://<bucket>/<object>.')
        return match.group(1), match.group(2)

    def start_bundle(self):
        self.credentials = sa_service.Credentials.from_service_account_info(
            self._sa)
        self.gcs_client = storage.Client(credentials=self.credentials)

    def process(self, element, *args, **kwargs):
        """Parse the blob and download as bytestring.

        Args:
            element:
            *args:
            **kwargs:
        """
        bucket_name, blob_name = self.parse_gcs_path(element)
        bucket = self.gcs_client.bucket(bucket_name)
        blob = bucket.blob(blob_name)
        blob_as_string = blob.download_as_string(start=0,
                                                 end=self.max_header_size
                                                 ).decode("utf8")

        first_row = blob_as_string.splitlines()[0]
        return first_row.strip().split(',')


def parse_columns(csv_file):
    """
    Args:
        csv_file:
    """
    return csv_file.readline().strip().split(',')


def create_bq_dict(line, column_names):
    """
    Args:
        line:
        column_names:
    """
    return dict(zip(column_names, line))


@beam.ptransform_fn
@beam.typehints.with_input_types(beam.Pipeline)
@beam.typehints.with_output_types(beam.typehints.Dict[Text, Any])
def ReadFilesFromGCS(pipeline: beam.Pipeline,
                     source_args: Dict[Text, Any]) -> beam.pvalue.PCollection:

    """
    Args:
        pipeline (beam.Pipeline):
        source_args:
    """
    gcs_path = source_args['base_path']
    bucket_re = re.match('^gs://([^/]+)', gcs_path)
    bucket = bucket_re.group(1)
    if bucket_re is None:
        raise ValueError(
            'GCS base path must be in the form gs://<bucket>/.')

    sa = source_args['service_account']

    credentials = sa_service.Credentials.from_service_account_info(sa)
    client = storage.Client(credentials=credentials)

    blob_iter = client.list_blobs(bucket_or_name=bucket)
    blob_list = list(blob_iter)

    uri_list = list(os.path.join(gcs_path, blob.name) for blob in blob_list)

    allowed_file_exts = [".csv", ".txt"]  # ".dat"

    # weed out bad file exts with this logic
    csv_files = [uri for uri in uri_list if os.path.splitext(uri)[1]
                 in allowed_file_exts]

    if not csv_files:
        raise RuntimeError("Error: No files found with extensions {0}. "
                           "".format(allowed_file_exts))

    column_names = (pipeline
                    | "CreateCSVPColl" >> beam.Create(csv_files[:1])
                    | "InferColumns" >> beam.ParDo(
                CSVColumnInferrer(service_account=sa))
                    )

    colnames = beam.pvalue.AsList(column_names)
    # column_names = io_utils.load_csv_column_names(csv_files[0])
    # for csv_file in csv_files[1:]:
    #     if io_utils.load_csv_column_names(csv_file) != column_names:
    #         raise RuntimeError(
    #             'Files in same split {} have different header.'.format(
    #                 file_pattern))

    parsed_csv_lines = (
            pipeline
            | "CreateCSVs" >> beam.Create(csv_files)
            | "ReadCSVBlobs" >> beam.ParDo(ProduceGCSBlobs(service_account=sa))
            | 'ReadFromText' >> beam.io.ReadAllFromText(skip_header_lines=1)
            | 'ParseCSVLine' >> beam.ParDo(
        csv_decoder.ParseCSVLine(delimiter=','))
            | 'ExtractParsedCSVLines' >> beam.Map(create_bq_dict, colnames)
            )

    return parsed_csv_lines


class GCSCSVSource(SourceInterface):
    def get_config_spec(self):
        return {'base_path': ExecutionParameter(type=Text),
                'service_account': ExecutionParameter(type=Dict[Text, Any])}

    def config_parser(self, config):
        """
        Args:
            config:
        """
        return config

    def beam_transform(self):
        return ReadFilesFromGCS



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import tensorflow as tf
import yaml
from tfx.dsl.components.base.base_executor import BaseExecutor
from tfx import types

from zenml.core.components.split_gen.executor import Executor

"""Test for the custom bigquery_pusher executor"""


class ExecutorTest(tf.test.TestCase):

    def setUp(self):
        """
        Setting up the input dict, output dict and exec parameters
        """
        self.input_dict = {}

        with open('/home/hamza/workspace/maiot/gdp/config.yaml') as f:
            experiment_dict = yaml.load(f)['experiment']
        group = experiment_dict['asset_groups']['horizontal_group_1']

        # Create execution parameters
        _bq_args = {
            'project': 'maiot-machine-learning',
            'dataset': 'adac_kasasi',
            'table': 'tfx_test'
        }

        self.exec_properties = {
            'bq_args': _bq_args,
            'group_yaml': group
        }

        # TODO: [LOW] To be updated
        self.output_dict = {
            'examples': [types.Artifact('ExamplesPath', split=split_name)
            for split_name in ['train', 'eval']]
        }
        # Creating a context for the executors to run the tests locally
        # TODO: Does not work with dataflow somehow
        options = [
            '--job_name=split_gen_executor_test'
            '--runner=DataflowRunner',
            '--project=maiot-machine-learning',
            '--temp_location=gs://adac_kasasi/tfx/split_gen/test/'
        ]
        self.context = BaseExecutor.Context(options)

    def testDo(self):
        # Create the executor with the custom context and run it with the
        # test input dict, output dict and exec properties
        split_gen_executor = Executor(self.context)
        split_gen_executor.Do(self.input_dict,
                              self.output_dict,
                              self.exec_properties)


if __name__ == '__main__':
    tf.test.main()



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import logging
from typing import Text, Type

from tfx.utils.import_utils import import_class_by_path

from zenml.core.repo.git_wrapper import GitWrapper
from zenml.core.repo.repo import Repository
from zenml.utils.constants import APP_NAME
from zenml.utils.version import __version__


def create_zenml_pin():
    """Creates a ZenML pin for source pinning from release version."""
    return f'{APP_NAME}_{__version__}'


def resolve_standard_source_path(source_path: Text) -> Text:
    """Creates a ZenML pin for source pinning from release version."""
    pin = create_zenml_pin()
    return f'{source_path}@{pin}'


def is_standard_pin(pin: Text) -> bool:
    """
    Returns True if pin is valid ZenML pin, else False.

    Args:
        pin (str): potential ZenML pin like 'zenml_0.1.1'
    """
    if pin.startswith(f'{APP_NAME}_'):
        return True
    return False


def is_standard_step(source_path: Text) -> bool:
    """
    Returns True if source_path is a standard ZenML class.

    Args:
        source_path (str): relative module path e.g. this.module.Class
    """
    if source_path.split('.')[0] == 'zenml':
        return True
    return False


def get_path_from_source(source_path):
    """
    Get file path from source

    Args:
        source_path (str): relative module path e.g. this.module.Class
    """
    file_path = "/".join(source_path.split('.')[:-1]) + '.py'
    return file_path


def resolve_source_path(source_path: Text) -> Text:
    """
    Resolves source path with an optional sha using Git.

    Args:
        source_path (str): relative module path e.g. this.module.Class
    """
    if is_standard_step(source_path):
        # that means use standard version
        return resolve_standard_source_path(source_path)

    # otherwise use Git resolution
    r: Repository = Repository.get_instance()
    wrapper: GitWrapper = r.get_git_wrapper()
    source_path = wrapper.resolve_source_path(source_path)
    return source_path


def load_source_path_class(source_path: Text) -> Type:
    """
    Loads a Python class from the path provided.

    Args:
        source_path (str): relative module path e.g. this.module.Class[@sha]
    """
    r: Repository = Repository.get_instance()
    pin = source_path.split('@')[-1]
    source = source_path.split('@')[0]
    is_standard = is_standard_pin(pin)

    if '@' in source_path and not is_standard:
        logging.info('Pinned step found with git sha. '
                     'Loading class from git history.')
        wrapper: GitWrapper = r.get_git_wrapper()

        # Get reference to current sha
        active_branch = wrapper.git_repo.active_branch.name

        # critical step
        logging.warning('Found source with a pinned sha. Will now stash '
                        'current changes and keep executing.')
        try:
            wrapper.stash()
        except Exception:
            # we need to make sure that an unsafe operation did not happen
            # TODO: [HIGH] Implement proper exception handling here.
            raise Exception

        try:
            wrapper.checkout(pin)
        except Exception:
            # popping needs to happen in any case
            wrapper.checkout(active_branch)
            wrapper.stash_pop()
            raise Exception

        class_ = import_class_by_path(source)
        wrapper.checkout(active_branch)
        wrapper.stash_pop()
    elif '@' in source_path and is_standard:
        logging.info(f'Default {APP_NAME} class used. Loading directly.')
        # TODO: [LOW] Check if ZenML version is installed before loading.
        class_ = import_class_by_path(source)
    else:
        logging.info('Unpinned step found with no git sha. Attempting to '
                     'load class from current repository state.')
        class_ = import_class_by_path(source)

    return class_



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import bisect
import logging
import os
from typing import Dict, Text, Any, List

import apache_beam as beam
import tensorflow as tf
from google.cloud import bigquery

from zenml.core.components.split_gen import constants
from zenml.core.components.split_gen.utils import get_categorical_value
from zenml.core.components.split_gen.utils_legacy import map_cat_to_fold


class BigQueryConverter(object):
    """Helper class for bigquery result row to tf example conversion."""

    def __init__(self, bq_args):
        client = bigquery.Client(bq_args['project'])
        table_ref = client.dataset(bq_args['dataset']).table(bq_args['table'])
        table = client.get_table(table_ref)
        schema = table.schema
        self._type_map = {s.name: s.field_type for s in schema}

    def RowToExample(self, instance: Dict[Text, Any]) -> tf.train.Example:
        """Convert bigquery result row to tf example."""
        feature = {}
        for key, value in instance.items():
            # we can safely ignore row_number now
            if key == constants.DEFAULT_ROW_NUMBER:
                continue

            data_type = self._type_map[key]
            kwargs = {}
            if data_type == 'INTEGER':
                if value is not None:
                    kwargs = {'value': [value]}
                feature[key] = tf.train.Feature(
                    int64_list=tf.train.Int64List(**kwargs))
            elif data_type == 'FLOAT':
                if value is not None:
                    kwargs = {'value': [value]}
                feature[key] = tf.train.Feature(
                    float_list=tf.train.FloatList(**kwargs))
            elif data_type == 'STRING':
                if value is not None:
                    kwargs = {'value': [tf.compat.as_bytes(value)]}
                feature[key] = tf.train.Feature(
                    bytes_list=tf.train.BytesList(**kwargs))
            elif data_type == 'TIMESTAMP' or data_type == 'DATETIME':
                if value is not None:
                    kwargs = {'value': [tf.compat.as_bytes(str(value))]}
                feature[key] = tf.train.Feature(
                    bytes_list=tf.train.BytesList(**kwargs))
                pass
            elif data_type == 'BOOLEAN':
                if value is not None:
                    kwargs = {'value': [int(value)]}
                feature[key] = tf.train.Feature(
                    int64_list=tf.train.Int64List(**kwargs))
            else:
                raise RuntimeError(
                    'BigQuery column type {} is not supported.'.format(
                        data_type))
        return tf.train.Example(features=tf.train.Features(feature=feature))


class CatDictCombiner(beam.CombineFn):
    def create_accumulator(self, **kwargs) -> Dict[Text, int]:
        return dict()

    def add_input(self, accumulator: Dict[Text, int],
                  element: Dict[Text, int],
                  **kwargs,
                  ) -> Dict[Text, int]:
        k, v = element
        if k not in accumulator:
            accumulator[k] = v
        return accumulator

    def merge_accumulators(self,
                           accumulators: List[Dict[Text, int]],
                           **kwargs,
                           ) -> Dict[Text, int]:
        d = {}
        for acc in accumulators:
            for k, v in acc.items():
                if k not in d:
                    d[k] = v
        return d

    def extract_output(self,
                       accumulator: Dict[Text, Any],
                       **kwargs) -> Dict[Text, int]:
        accumulator["total"] = sum(accumulator.values())
        return accumulator


@beam.ptransform_fn
@beam.typehints.with_input_types(beam.Pipeline)
@beam.typehints.with_output_types(beam.typehints.Dict[Text, Any])
def ReadFromBigQuery(pipeline: beam.Pipeline,
                     query: Text,
                     exec_properties: Dict) -> beam.pvalue.PCollection:
    logging.info('Executing query: {}'.format(query))
    return (pipeline
            | 'QueryTable' >> beam.io.Read(
                beam.io.BigQuerySource(
                    project=exec_properties[constants.BQ_PROJECT],
                    query=query,
                    use_standard_sql=True)))


def SplitPartitionFn(element: tf.train.Example,
                     num_partitions: int,
                     exec_properties: Dict[Text, any],
                     category_map: Dict[Text, any],
                     limit_map: Dict[Text, any]) -> int:
    element_category = constants.DEFAULT_CATEGORY

    example, row_number = element

    # splitting by category (horizontal)
    if constants.CAT_COL in exec_properties:
        cat_col = exec_properties[constants.CAT_COL]

        element_category = get_categorical_value(example, cat_col=cat_col)
        # map_cat_to_fold returns the fold ID or -1, which is a placeholder
        # for "does not belong into any fold"
        fold_id = map_cat_to_fold(element_category, category_map)

        if exec_properties[constants.IDX_COL] is not None:
            if fold_id > 0:
                # fold_id > 0 <=> fold is eval or test, so return it,
                # otherwise split the train set further (hybrid split)
                return fold_id
        else:
            if fold_id >= 0:
                # return fold
                # TODO: What happens in the nosplit scenario?
                #  Does it even make sense to allow a categorical nosplit?
                return fold_id

    # move on to a hybrid split
    # hybrid splitting also happens implicitly when the element has a
    # CAT_COL feature category that was not selected into any of the folds.

    # splitting by row number (vertical)
    # element_row_number = element[constants.DEFAULT_ROW_NUMBER]

    # the row number limits for train, test, eval split by category
    row_num_limits = list(limit_map[element_category].values())

    # Python function to put row numbers into buckets
    return bisect.bisect(row_num_limits, row_number)


class CountExamples(beam.PTransform):
    def __init__(self):
        super(CountExamples, self).__init__()
        self._counter = 0

    def increase_count(self, element, *args, **kwargs):
        self._counter += 1
        return element, self._counter

    def expand(self, pcoll):
        return (pcoll
                | "CountExamples" >> beam.Map(self.increase_count))


@beam.ptransform_fn
@beam.typehints.with_input_types(bytes)
@beam.typehints.with_output_types(beam.pvalue.PDone)
def WriteSplit(example_split: beam.pvalue.PCollection,
               output_split_path: Text) -> beam.pvalue.PDone:
    """Shuffles and writes output split."""
    out_dir = os.path.join(output_split_path, 'data_tfrecord')
    logging.info('Writing output TFRecords to {}'.format(
        out_dir
    ))
    return (example_split
            | 'Write' >> beam.io.WriteToTFRecord(
                out_dir,
                file_name_suffix='.gz'))



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from zenml.core.pipelines.config.config_utils import ConfigKeys


class GlobalKeys(ConfigKeys):
    VERSION = 'version'
    SPECS = 'specs'
    STEPS = 'steps'


class EnvironmentKeys(ConfigKeys):
    EXPERIMENT_NAME = 'experiment_name'
    ARTIFACT_STORE = 'artifact_store'
    METADATA_STORE = 'metadata_store'
    ENABLE_CACHE = 'enable_cache'
    BACKENDS = 'backends'


class BackendKeys(ConfigKeys):
    TYPE = 'type'
    ARGS = 'args'


class StepKeys(ConfigKeys):
    FN = 'fn'
    ARGS = 'args'


class TrainingSteps(ConfigKeys):
    DATA = 'data'
    SPLIT = 'split'
    SEQUENCING = 'sequencing'
    PREPROCESSING = 'preprocessing'
    TRAINING = 'training'
    EVALUATION = 'evaluation'
    DEPLOYMENT = 'deployment'


class DataSteps(ConfigKeys):
    DATA = 'data'

# class AIPlatformKeys(ConfigKeys):
#     SCALE_TIER = 'scale_tier'
#     RUNTIME_VERSION_ = 'runtime_version'
#     PYTHON_VERSION_ = 'python_version'
#     MAX_RUNNING_TIME_ = 'max_running_time'
#
#
# class OrchestratorKeys(ConfigKeys):
#     TYPE = 'type'
#     ARGS = 'args'
#
#
# class BigQueryKeys(ConfigKeys):
#     PROJECT_ = 'project'
#     DATASET = 'dataset'
#     TABLE_ = 'table'
#
#
# class KubeflowKeys(ConfigKeys):
#     IMAGE_ = ''
#
#
# class BeamKeys(ConfigKeys):
#     RUNNER = 'runner'
#     WORKER_MACHINE_TYPE_ = 'worker_machine_type'
#     NUM_WORKERS_ = 'num_workers'
#     MAX_NUM_WORKERS_ = 'max_num_workers'
#     DISK_SIZE_GB_ = 'disk_size_gb'
#     AUTOSCALING_ALGORITHM_ = 'autoscaling_algorithm'
#
#
# class GCPOrchestratorKeys(ConfigKeys):
#     SERVICE_ACCOUNT = 'service_account'
#     PROJECT_ = 'project'
#     ZONE_ = 'zone'
#     NAME_ = 'name'
#     IMAGE_NAME_ = 'image_name'
#     PREEMPTIBLE_ = 'preemptible'
#     MACHINE_TYPE_ = 'machine_type'
#
#
# class GCPProviderKeys(ConfigKeys):
#     ARTIFACT_STORE_ = 'artifact_store'
#
#
# class MetadataKeys(ConfigKeys):
#     TYPE = 'type'
#     HOST_ = 'host'
#     PORT_ = 'port'
#     DATABASE_ = 'database'
#     USERNAME_ = 'username'
#     PASSWORD_ = 'password'
#



# #  Copyright (c) maiot GmbH 2020. All Rights Reserved.
# #
# #  Licensed under the Apache License, Version 2.0 (the "License");
# #  you may not use this file except in compliance with the License.
# #  You may obtain a copy of the License at:
# #
# #       http://www.apache.org/licenses/LICENSE-2.0
# #
# #  Unless required by applicable law or agreed to in writing, software
# #  distributed under the License is distributed on an "AS IS" BASIS,
# #  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# #  or implied. See the License for the specific language governing
# #  permissions and limitations under the License.
#
# from typing import Text, Dict, Any
#
# import apache_beam as beam
# import oauth2client.service_account
# from apache_beam.internal.http_client import get_new_http
# from apache_beam.io.gcp import bigquery_tools
# from apache_beam.io.gcp.internal.clients.bigquery import BigqueryV2
#
# from zenml.core.components.data_gen.constants import BQArgs
# from zenml.core.components.data_gen.sources.interface import SourceInterface
#
#
# class CEBigQuerySource(beam.io.BigQuerySource):
#     def __init__(self, credentials, *args, **kwargs):
#         """
#         Args:
#             credentials:
#             *args:
#             **kwargs:
#         """
#         super(CEBigQuerySource, self).__init__(*args, **kwargs)
#         # self.bq_client = bq_client
#         self.credentials = credentials
#         self.bq_client = BigqueryV2(
#             http=get_new_http(),
#             credentials=self.credentials,
#             response_encoding='utf8')
#
#     def reader(self, *args, **kwargs):
#         """
#         Args:
#             *args:
#             **kwargs:
#         """
#         return bigquery_tools.BigQueryReader(
#             source=self,
#             test_bigquery_client=self.bq_client,
#             use_legacy_sql=self.use_legacy_sql,
#             flatten_results=self.flatten_results,
#             kms_key=self.kms_key)
#
#
# @beam.ptransform_fn
# @beam.typehints.with_input_types(beam.Pipeline)
# @beam.typehints.with_output_types(beam.typehints.Dict[Text, Any])
# def ReadFromBigQuery(pipeline: beam.Pipeline,
#                      source_args: Dict[Text, Any]) -> beam.pvalue.PCollection:
#     """
#     Args:
#         pipeline (beam.Pipeline):
#         source_args:
#     """
#     query = '''SELECT * FROM `{project}.{dataset}.{table}`'''.format(
#         project=source_args[BQArgs.PROJECT],
#         dataset=source_args[BQArgs.DATASET],
#         table=source_args[BQArgs.TABLE])
#
#     if BQArgs.LIMIT_ in source_args:
#         query += 'LIMIT {}'.format(source_args[BQArgs.LIMIT_])
#
#     if BQArgs.SERVICE_ACCOUNT_ in source_args:
#         # Its a private dataset
#         sa = source_args[BQArgs.SERVICE_ACCOUNT_]
#         scope = "https://www.googleapis.com/auth/bigquery"
#         credentials = oauth2client.service_account.ServiceAccountCredentials. \
#             from_json_keyfile_dict(keyfile_dict=sa,
#                                    scopes=scope)
#         bq_source = CEBigQuerySource(
#             # bq_client=bq_client,
#             credentials=credentials,
#             project=source_args[BQArgs.PROJECT],
#             query=query,
#             use_standard_sql=True)
#
#     else:  # we have a public table, no special permissions needed
#         bq_source = beam.io.BigQuerySource(
#             project=source_args[BQArgs.PROJECT],
#             query=query,
#             use_standard_sql=True)
#
#     return (pipeline
#             | 'QueryTable' >> beam.io.Read(source=bq_source))
#
#
# class BigQuerySource(SourceInterface):
#     ARG_KEYS = BQArgs
#
#     def beam_transform(self):
#         return ReadFromBigQuery



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the 'License');
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an 'AS IS' BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from zenml.core.pipelines.config.config_utils import ConfigKeys

DATA_SPLIT_NAME = 'examples'
DATA_COL = '_data_'
SCHEMA = 'schema'

# Possible types and sources (with the corresponding args)

TYPE_TEXT = 'text'
TYPE_TABULAR = 'tabular'
TYPE_IMAGE = 'image'
TYPE_IMAGE_SEG = 'image_segmentation'

SOURCE_BQ = 'bq'
SOURCE_GCS = 'gcs'
SOURCE_MVG_BQ = 'mvg_bq'
SOURCE_LOCAL_IMAGE = 'local_image'
SOURCE_LOCAL_CSV = 'local_csv'
SOURCE_GCS_IMAGE = 'gcs_image'
SOURCE_GCS_CSV = 'gcs_csv'

# local storage data source args
BINARY_DATA = 'binary_data'
METADATA = 'metadata'
FILE_EXT = 'file_ext'
FILE_NAME = 'file_name'
BASE_PATH = 'base_path'
FILE_PATH = 'file_path'

# image data constants
IMAGE = 'image'
LABEL = 'label'
MASK = 'mask'


class BQArgs(ConfigKeys):
    PROJECT = 'project'
    DATASET = 'dataset'
    TABLE = 'table'
    LIMIT_ = 'limit'
    SERVICE_ACCOUNT_ = 'service_account'


class LocalImageArgs(ConfigKeys):
    BASE_PATH = 'base_path'


class LocalCSVArgs(ConfigKeys):
    BASE_PATH = 'base_path'


class MVGBQArgs(ConfigKeys):
    PROJECT = 'project'
    DATASET = 'dataset'
    TABLE = 'table'
    ASSET_IDS_ = 'asset_ids'
    LIMIT_ = 'limit'
    PGN_DATABASE_ = 'pgn_database'
    ERROR_DATABASE_ = 'error_database'


class GCSArgs(ConfigKeys):
    SERVICE_ACCOUNT = 'service_account'
    PATH = 'path'


class GCSImageArgs(ConfigKeys):
    BASE_PATH = 'base_path'
    SERVICE_ACCOUNT = 'service_account'


# Component keys
class SpecParamKeys(ConfigKeys):
    SOURCE = 'source'
    SOURCE_ARGS = 'args'


class DatasourceKeys(ConfigKeys):
    SOURCE = 'source'
    TYPE = 'type'
    ARGS = 'args'


class DestinationKeys(ConfigKeys):
    PROJECT = 'project'
    DATASET = 'dataset'
    TABLE = 'table'



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Orchestrator for simple GCP VM backend"""

import base64
import json
import os
import time
from typing import Dict, Any
from typing import Text

import googleapiclient.discovery
from google.oauth2 import service_account as sa

from zenml.core.backends.orchestrator.local.orchestrator_local_backend import \
    OrchestratorLocalBackend


class OrchestratorGCPBackend(OrchestratorLocalBackend):
    """Orchestrates pipeline in a VM.

    Every ZenML pipeline runs in backends.
    """
    BACKEND_TYPE = 'gcp'
    SOURCE_DISK_IMAGE = "projects/cos-cloud/global/images/cos-stable-81" \
                        "-12871-1196-0"

    def __init__(self,
                 project,
                 image: Text = None,
                 zone: Text = 'europe-west1-b',
                 instance_name: Text = str(int(time.time())),
                 machine_type: Text = 'n1-standard-4',
                 preemptible: bool = True,
                 service_account: Text = None,
                 **unused_kwargs):
        self.project = project
        self.zone = zone
        self.image = image
        self.machine_type = machine_type
        self.instance_name = instance_name
        self.preemptible = preemptible

        if service_account:
            scopes = ['https://www.googleapis.com/auth/cloud-platform']
            self.credentials = \
                sa.Credentials.from_service_account_info(
                    sa, scopes=scopes)
        else:
            self.credentials = None

        super().__init__(**unused_kwargs)
        raise NotImplementedError('Its coming soon!')

    def run(self, config: Dict[Text, Any]):
        """
        This run function essentially calls an underlying TFX orchestrator run.
        However it is meant as a higher level abstraction with some
        opinionated decisions taken.

        Args:
            config: a ZenML config dict
        """
        # Instantiate google compute service
        compute = googleapiclient.discovery.build('compute', 'v1',
                                                  credentials=self.credentials)

        # Configure the machine
        machine_type = f"zones/{self.zone}/machineTypes/{self.machine_type}"
        startup_script = open(
            os.path.join(
                os.path.dirname(__file__), 'startup-script.sh'), 'r').read()

        config_encoded = base64.b64encode(json.dumps(config).encode())
        c_params = f'run_pipeline --config_b64 {config_encoded}'

        compute_config = {
            'name': self.instance_name,
            'zone': f'projects/{self.project}/zones/{self.zone}',
            'machineType': machine_type,

            # Specify if preemtible
            'scheduling': {'preemptible': self.preemptible},

            # Specify the boot disk and the image to use as a source.
            'disks': [
                {
                    'boot': True,
                    'autoDelete': True,
                    'initializeParams': {
                        'sourceImage': self.SOURCE_DISK_IMAGE,
                        'diskSizeGb': '100'
                    }
                }
            ],

            # Specify a network interface with NAT to access the public
            # internet.
            'networkInterfaces': [{
                'network': 'global/networks/default',
                'accessConfigs': [
                    {'type': 'ONE_TO_ONE_NAT', 'name': 'External NAT'}
                ]
            }],

            # Allow the instance to access cloud storage and logging.
            'serviceAccounts': [{
                'email': 'default',
                'scopes': [
                    # to be refined
                    'https://www.googleapis.com/auth/cloud-platform',
                    # 'https://www.googleapis.com/auth/devstorage.read_write',
                    # 'https://www.googleapis.com/auth/logging.write'
                ]
            }],

            # Metadata is readable from the instance and allows you to
            # pass configuration from deployment scripts to instances.
            'metadata': {
                'items': [
                    {
                        # Startup script is automatically executed by the
                        # instance upon startup.
                        'key': 'startup-script',
                        'value': startup_script
                    },
                    {
                        'key': 'image_name',
                        'value': self.image,
                    },
                    {
                        'key': 'container_params',
                        'value': c_params,
                    },
                ]
            }
        }

        try:
            return compute.instances().insert(
                project=self.project,
                zone=self.zone,
                body=compute_config).execute()
        except Exception as e:
            raise AssertionError(f"GCP VM failed to launch with the following "
                                 f"error: {str(e)}")



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base High-level ZenML Pipeline definition"""

import logging
import os
from abc import abstractmethod
from typing import Dict, Text, Any, Optional, List
from uuid import uuid4

from tfx.utils import io_utils

from zenml.core.backends.base_backend import BaseBackend
from zenml.core.backends.orchestrator.local.orchestrator_local_backend import \
    OrchestratorLocalBackend
from zenml.core.datasources.base_datasource import BaseDatasource
from zenml.core.metadata.metadata_wrapper import ZenMLMetadataStore
from zenml.core.pipelines.utils import parse_yaml_beam_args
from zenml.core.repo.artifact_store import ArtifactStore
from zenml.core.repo.repo import Repository
from zenml.core.repo.zenml_config import METADATA_KEY
from zenml.core.standards import standard_config as keys
from zenml.core.steps.base_step import BaseStep
from zenml.utils.constants import CONFIG_VERSION, APP_NAME
from zenml.utils.print_utils import to_pretty_string, PrintStyles
from zenml.utils.version import __version__
from zenml.utils.zenml_analytics import track, CREATE_PIPELINE, RUN_PIPELINE


class BasePipeline:
    """Base class for all ZenML pipelines.

    Every ZenML pipeline should override this class.
    """
    PIPELINE_TYPE = 'base'

    @track(event=CREATE_PIPELINE)
    def __init__(self,
                 name: Text,
                 enable_cache: Optional[bool] = True,
                 steps_dict: Dict[Text, BaseStep] = None,
                 backends_dict: Dict[Text, BaseBackend] = None,
                 metadata_store: Optional[ZenMLMetadataStore] = None,
                 artifact_store: Optional[ArtifactStore] = None,
                 datasource: Optional[BaseDatasource] = None,
                 pipeline_name: Optional[Text] = None,
                 *args, **kwargs):
        """
        Construct a base pipeline.

        Args:
            name: outwards-facing name of the pipeline
            pipeline_name: unique name that identifies the pipeline after
            its run
            enable_cache: boolean indicator whether caching should be used
            steps_dict: optional dict of steps
            backends_dict: optional dict of backends
            metadata_store: chosen metadata store, if None use default
            artifact_store: chosen artifact store, if None use default
        """
        self.repo: Repository = Repository.get_instance()
        self.name = name
        if pipeline_name:
            self.pipeline_name = pipeline_name
            self.file_name = self.pipeline_name + '.yaml'
        else:
            # if pipeline_name is not None then its not a new pipeline
            self.pipeline_name = self.create_pipeline_name_from_name()
            self.file_name = self.pipeline_name + '.yaml'
            # check duplicates here as its a 'new' pipeline
            if self.file_name in self.repo.get_pipeline_file_paths(
                    only_file_names=True):
                raise AssertionError(f'Pipeline names must be unique in the '
                                     f'repository. There is already a '
                                     f'pipeline '
                                     f'called {self.name}')

        self.enable_cache = enable_cache

        if steps_dict is None:
            self.steps_dict: Dict[Text, BaseStep] = {}
        else:
            self.steps_dict = steps_dict

        # Backends
        if backends_dict is None:
            self.backends_dict: Dict[Text, BaseBackend] = \
                self.get_default_backends()
        else:
            self.backends_dict = backends_dict

        # Metadata store
        if metadata_store:
            self.metadata_store = metadata_store
        else:
            # use default
            self.metadata_store = self.repo.get_metadata_store()

        # Artifact store
        if artifact_store:
            self.metadata_store = metadata_store
        else:
            # use default
            self.artifact_store = self.repo.get_artifact_store()

        # Datasource
        if datasource:
            self.datasource = datasource
        else:
            self.datasource = None

    def __str__(self):
        return to_pretty_string(self.to_config())

    def __repr__(self):
        return to_pretty_string(self.to_config(), style=PrintStyles.PPRINT)

    @abstractmethod
    def get_tfx_component_list(self, config: Dict[Text, Any]) -> List:
        """
        Converts config to TFX components list. This is the point in the
        framework where ZenML Steps get translated into TFX pipelines.

        Args:
            config: dict of ZenML config.
        """
        pass

    @staticmethod
    def get_type_from_file_name(file_name: Text):
        """
        Gets type of pipeline from file name.

        Args:
            file_name: YAML file name of pipeline.
        """
        return file_name.replace('.yaml', "").split('_')[0]

    @staticmethod
    def get_type_from_pipeline_name(pipeline_name: Text):
        """
        Gets type from pipeline name.

        Args:
            pipeline_name (str): simple string name.
        """
        return pipeline_name.split('_')[0]

    @staticmethod
    def get_name_from_pipeline_name(pipeline_name: Text):
        """
        Gets name from pipeline name.

        Args:
            pipeline_name (str): simple string name.
        """
        return pipeline_name.split('_')[1]

    @staticmethod
    def get_pipeline_spec(pipeline_dict: Dict[Text, Any]):
        env_dict = pipeline_dict[keys.GlobalKeys.ENV]

        # Dataset, experiment
        # TODO: [LOW] Check if using the artifact store ID makes sense
        #  It is used currently to make all pipelines have the same base
        #  context in ML metadata store.
        repo_id = \
            Repository.get_instance().get_artifact_store().unique_id
        pipeline_name = env_dict[keys.EnvironmentKeys.EXPERIMENT_NAME]

        # Artifact- and metadata store
        artifact_store = env_dict[keys.EnvironmentKeys.ARTIFACT_STORE]
        if not artifact_store.startswith('gs://'):
            artifact_store = os.path.abspath(artifact_store)

        metadata_store: ZenMLMetadataStore = ZenMLMetadataStore.from_config(
            env_dict[keys.EnvironmentKeys.METADATA_STORE])
        metadata_connection_config = metadata_store.get_tfx_metadata_config()

        # Pipeline settings
        pipeline_enable_cache = env_dict[keys.EnvironmentKeys.ENABLE_CACHE]
        pipeline_root = os.path.join(artifact_store, repo_id)
        pipeline_log = os.path.join(pipeline_root, 'logs', pipeline_name)
        pipeline_temp = os.path.join(pipeline_root, 'tmp', pipeline_name)

        # Execution
        # TODO: [MEDIUM] Refactor this into execution backends?
        execution = env_dict[keys.EnvironmentKeys.BACKENDS]['orchestrator']

        exec_type = execution[keys.BackendKeys.TYPE]
        exec_args = execution[keys.BackendKeys.ARGS]

        if exec_type == 'beam':
            dist_path = os.path.join(os.getcwd(), 'dist')
            req_path = os.path.join(dist_path, 'requirements.txt')
            io_utils.write_string_file(req_path, '')
            gz_path = os.path.join(dist_path, '{}-{}.tar.gz'.format(
                APP_NAME, __version__))

            exec_args['extra_package'] = gz_path
            exec_args['requirements_file'] = req_path
            exec_args['setup_file'] = os.path.join(os.getcwd(), 'setup.py')
            exec_args['job_name'] = 'gdp-' + pipeline_name
            exec_args['temp_location'] = pipeline_temp
            exec_args['staging_location'] = pipeline_temp
            exec_args = parse_yaml_beam_args(exec_args)
        elif exec_type == 'local':
            pass
        else:
            raise AssertionError('Unknown orchestrator!')

        # Training
        # TODO: [MEDIUM] Refactor this into training backends?
        if 'training' in env_dict[keys.EnvironmentKeys.BACKENDS]:
            training = env_dict[keys.EnvironmentKeys.BACKENDS]['training']

            training_type = training[keys.BackendKeys.TYPE]
            training_args = training[keys.BackendKeys.ARGS]

            train_args = dict()
            if training_type == 'gcaip':
                train_args = {'project': training_args['project_id'],
                              'region': training_args['gcp_region'],
                              'jobDir': pipeline_temp,
                              # TODO: where is the image?
                              # 'masterConfig': {'imageUri': TRAINER_IMAGE},
                              'scaleTier': training_args['scale_tier']}

                if 'runtime_version' in training_args:
                    train_args['runtimeVersion'] = training_args[
                        'runtime_version']

                if 'python_version' in training_args:
                    train_args['pythonVersion'] = training_args[
                        'python_version']

                if 'max_running_time' in training_args:
                    train_args['scheduling'] = {
                        'maxRunningTime': training_args['max_running_time']}

            training_dict = {
                'type': training_type,
                'args': training_args,
            }
        else:
            training_dict = {}

        return {'repo_id': repo_id,
                'pipeline_root': pipeline_root,
                keys.EnvironmentKeys.ENABLE_CACHE: pipeline_enable_cache,
                'pipeline_log_root': pipeline_log,
                'metadata_connection_config': metadata_connection_config,
                'execution_args': exec_args,
                'training_args': training_dict}

    @classmethod
    def from_config(cls, config: Dict):
        """
        Convert from pipeline config to ZenML Pipeline object.

        All steps are also populated and configuration set to parameters set
        in the config file.

        Args:
            config: a ZenML config in dict-form (probably loaded from YAML).
        """
        # populate steps
        steps_dict: Dict = {}
        for step_key, step_config in config[keys.GlobalKeys.STEPS].items():
            steps_dict[step_key] = BaseStep.from_config(step_config)

        env = config[keys.GlobalKeys.ENV]
        pipeline_name = env[keys.EnvironmentKeys.EXPERIMENT_NAME]
        name = BasePipeline.get_name_from_pipeline_name(
            pipeline_name=pipeline_name)

        backends_dict: Dict = {}
        for backend_key, backend_config in env[
            keys.EnvironmentKeys.BACKENDS].items():
            backends_dict[backend_key] = BaseBackend.from_config(
                backend_config)

        artifact_store = ArtifactStore(
            env[keys.EnvironmentKeys.ARTIFACT_STORE])
        metadata_store = ZenMLMetadataStore.from_config(
            config=env[METADATA_KEY]
        )

        datasource = BaseDatasource.from_config(config)

        from zenml.core.pipelines.pipeline_factory import pipeline_factory
        pipeline_type = BasePipeline.get_type_from_pipeline_name(
            pipeline_name)
        class_ = pipeline_factory.get_pipeline_by_type(pipeline_type)

        return class_(name=name, pipeline_name=pipeline_name,
                      enable_cache=env[keys.EnvironmentKeys.ENABLE_CACHE],
                      steps_dict=steps_dict,
                      backends_dict=backends_dict,
                      artifact_store=artifact_store,
                      metadata_store=metadata_store,
                      datasource=datasource)

    def create_pipeline_name_from_name(self):
        """Creates a unique pipeline name from user-provided name."""
        return self.PIPELINE_TYPE.lower() + '_' + self.name + '_' + str(
            uuid4())

    def get_steps_config(self) -> Dict:
        """Convert Step classes to steps config dict."""
        steps_config = {}
        for step_key, step in self.steps_dict.items():
            steps_config[step_key] = step.to_config()

        return {keys.GlobalKeys.STEPS: steps_config}

    def get_environment(self) -> Dict:
        """Get environment as a dict."""
        backends_config_dict = {}
        for b in self.backends_dict.values():
            backends_config_dict.update(b.to_config())

        return {
            keys.EnvironmentKeys.EXPERIMENT_NAME: self.pipeline_name,
            keys.EnvironmentKeys.ENABLE_CACHE: self.enable_cache,
            keys.EnvironmentKeys.BACKENDS: backends_config_dict,
            keys.EnvironmentKeys.ARTIFACT_STORE: self.artifact_store.path,
            keys.EnvironmentKeys.METADATA_STORE:
                self.metadata_store.to_config()
        }

    def to_config(self) -> Dict:
        """Converts pipeline to ZenML config."""
        # Create a ZenML pipeline.config.yaml
        steps_config = self.get_steps_config()

        # Add env config to it
        environment = self.get_environment()

        steps_config.update({
            keys.GlobalKeys.ENV: environment,
            keys.GlobalKeys.VERSION: CONFIG_VERSION,
            keys.GlobalKeys.DATASOURCE: self.datasource.to_config()
        })
        return steps_config

    def get_status(self) -> Text:
        """Get status of pipeline."""
        store = self.repo.get_metadata_store()
        return store.get_pipeline_status(self.pipeline_name)

    def register_pipeline(self, config: Dict[Text, Any]):
        """
        Registers a pipeline in the artifact store as a YAML file.

        Args:
            config: dict representation of ZenML config.
        """
        self.repo.register_pipeline(file_name=self.file_name, config=config)

    def get_default_backends(self) -> Dict:
        """Gets list of default backends for this pipeline."""
        # For base class, orchestration is always necessary
        return {
            OrchestratorLocalBackend.BACKEND_KEY: OrchestratorLocalBackend()
        }

    def run_config(self, config: Dict[Text, Any]):
        """
        Gets TFX pipeline from config.

        Args:
            config: dict of ZenML config.
        """
        if OrchestratorLocalBackend.BACKEND_KEY not in self.backends_dict:
            raise AssertionError('Orchestrator backend missing!')
        orch_backend = self.backends_dict[OrchestratorLocalBackend.BACKEND_KEY]

        # TODO: [LOW] Make sure this is subclass of OrchestratorLocalBackend
        orch_backend.run(config)

    @abstractmethod
    def steps_completed(self) -> bool:
        """Returns True if all steps complete, else raises exception"""
        pass

    def add_datasource(self, datasource: BaseDatasource):
        """
        Add datasource to pipeline.

        Args:
            datasource: class of type BaseDatasource
        """
        self.datasource = datasource
        self.steps_dict[keys.TrainingSteps.DATA] = datasource.get_data_step()

    @track(event=RUN_PIPELINE)
    def run(self,
            backends: Optional[List[BaseBackend]] = None,
            metadata_store: Optional[ZenMLMetadataStore] = None,
            artifact_store: Optional[ArtifactStore] = None):
        """
        Run the pipeline associated with the datasource.

        Args:
            backends (list): list of backends to use for this
            metadata_store: chosen metadata store, if None use default
            artifact_store: chosen artifact store, if None use default
             """

        # TODO: [HIGH] Important, think about separating register and run
        #  and that way ask user to input name while registering pipeline.

        # Resolve default
        if metadata_store:
            logging.warning('Changing the metadata_store or artifact_store '
                            'might cause your pipelines to be '
                            'non-reproducible and non-comparable.')
            self.metadata_store = metadata_store

        if artifact_store:
            logging.warning('Changing the metadata_store or artifact_store '
                            'might cause your pipelines to be '
                            'non-reproducible and non-comparable.')
            self.metadata_store = metadata_store

        # Check if steps are complete
        self.steps_completed()

        # Resolve backends compatibility
        if backends is None:
            backends = []
        for b in backends:
            if b.BACKEND_KEY not in self.backends_dict:
                raise Exception(f'Backend {b} not supported!')
            self.backends_dict[b.BACKEND_KEY] = b

        # Get the config dict
        config = self.to_config()

        # Register in the repository
        self.register_pipeline(config)

        # Run the pipeline
        self.run_config(config)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import base64
import enum
import logging
import sys
from abc import ABC
from typing import Dict, List, Text, Any

import apache_beam as beam
import numpy as np
import tensorflow as tf
from dateutil.parser import parse
from google.cloud import bigquery

from zenml.core.components.data_gen import constants
from zenml.core.components.data_gen.constants import DestinationKeys


def is_date(string, fuzzy=False):
    """
    Return whether the string can be interpreted as a date.

    :param string: str, string to check for date
    :param fuzzy: bool, ignore unknown tokens in string if True
    """
    try:
        parse(tf.compat.as_text(string), fuzzy=fuzzy)
        return True
    except ValueError:
        return False
    except TypeError:
        return False


def is_bytes(data):
    """

    Parameters
    ----------
    data: Bytes-like, checking if it is a string

    Returns: bool, if data is bytes-like
    -------

    """
    return type(data) is bytes


def is_smaller_than(data_dict: Dict[Text, Any], max_size=10485760):
    """

    Parameters
    ----------
    data_dict: Beam data dict to be uploaded into BQ
    max_size: maximum datapoint size in bytes, default 10 MiB

    Returns
    -------
    bool: Whether the data dict is smaller or larger than max_size bytes.

    """
    return sys.getsizeof(data_dict) < max_size


class DataType(enum.IntEnum):
    UNKNOWN = -1
    INT = 0
    FLOAT = 1
    TIMESTAMP = 2
    STRING = 3
    BYTES = 4


BQ_MAPPING = {
    DataType.STRING: 'STRING',
    DataType.INT: 'INTEGER',
    DataType.FLOAT: 'FLOAT',
    DataType.TIMESTAMP: 'TIMESTAMP',
    DataType.BYTES: 'BYTES'
}

_INT64_MIN = np.iinfo(np.int64).min
_INT64_MAX = np.iinfo(np.int64).max


@beam.typehints.with_output_types(beam.typehints.Dict[Text, DataType])
class DtypeInferrer(beam.CombineFn, ABC):
    """A beam.CombineFn to infer data types"""

    @staticmethod
    def _infer_value_type(value) -> DataType:
        try:
            float(value)
        except ValueError:
            # TODO: Check for string here, if not return UNKNOWN
            if is_date(value):
                return DataType.TIMESTAMP
            elif is_bytes(value):
                return DataType.BYTES
            elif isinstance(value, str):
                return DataType.STRING
            else:
                return DataType.UNKNOWN

        except TypeError:
            return DataType.UNKNOWN
        else:
            if float(value).is_integer():
                return DataType.INT
            else:
                return DataType.FLOAT

    def create_accumulator(self, **kwargs) -> Dict[Text, DataType]:
        return {}

    def add_input(self,
                  accumulator: Dict[Text, DataType],
                  element: Dict[Text, bytes],
                  **kwargs,
                  ) -> Dict[Text, DataType]:

        for key, value in element.items():
            previous_type = accumulator.get(key, None)
            current_type = self._infer_value_type(value)

            if previous_type is None or current_type > previous_type:
                accumulator[key] = current_type
        return accumulator

    def merge_accumulators(self,
                           accumulators: List[Dict[Text, DataType]],
                           **kwargs,
                           ) -> Dict[Text, DataType]:

        result = {}
        for single_accumulator in accumulators:
            for feature_name, feature_type in single_accumulator.items():
                if feature_name not in result \
                        or feature_type > result[feature_name]:
                    result[feature_name] = feature_type
        return result

    def extract_output(self, accumulator: Dict[Text, Any], **kwargs) -> \
            Dict[Text, Text]:
        final_schema = {}
        for feature, dtype in accumulator.items():
            if dtype in BQ_MAPPING:
                final_schema[feature] = BQ_MAPPING[dtype]
            else:
                final_schema[feature] = 'STRING'
                logging.error(f'Feature {feature} has UNKNOWN dtype {dtype}')

        return final_schema


def append_tf_example(data: Dict[Text, Any],
                      schema: Dict[Text, Any]) -> tf.train.Example:
    """Add tf example to row"""
    feature = {}
    new_data = data.copy()  # copy of the data

    for key, value in data.items():
        data_type = schema[key]
        kwargs = {}
        if data_type == BQ_MAPPING[DataType.INT]:
            if value is not None:
                kwargs = {'value': [int(value)]}
            feature[key] = tf.train.Feature(
                int64_list=tf.train.Int64List(**kwargs))
        elif data_type == BQ_MAPPING[DataType.FLOAT]:
            if value is not None:
                kwargs = {'value': [float(value)]}
            feature[key] = tf.train.Feature(
                float_list=tf.train.FloatList(**kwargs))
        elif data_type == BQ_MAPPING[DataType.TIMESTAMP]:
            if value is not None:
                ts = parse(tf.compat.as_text(value), fuzzy=True)
                ts = ts.strftime('%Y-%m-%dT%H:%M:%S.%f %Z')
                new_data[key] = ts
                kwargs = {'value': [tf.compat.as_bytes(ts)]}
            feature[key] = tf.train.Feature(
                bytes_list=tf.train.BytesList(**kwargs))
        elif data_type == BQ_MAPPING[DataType.STRING]:
            if value is not None:
                kwargs = {'value': [tf.compat.as_bytes(str(value))]}
            feature[key] = tf.train.Feature(
                bytes_list=tf.train.BytesList(**kwargs))
        elif data_type == BQ_MAPPING[DataType.BYTES]:
            val = tf.compat.as_bytes(value)
            # val.replace('''''')
            if value is not None:
                kwargs = {'value': [val]}
            feature[key] = tf.train.Feature(
                bytes_list=tf.train.BytesList(**kwargs))
        else:
            raise RuntimeError(f'Unknown data type {data_type} in the schema '
                               f'for value {value}!')

    tf_example = tf.train.Example(features=tf.train.Features(feature=feature))
    return tf_example


def init_bq_table(project, dataset, table, schema):
    client = bigquery.Client(project=project)

    # check if dataset exists, if not create
    datasets = [d.dataset_id for d in client.list_datasets()]
    if dataset not in datasets:
        dataset = client.create_dataset(dataset)
    else:
        dataset = client.dataset(dataset)

    table_ref = dataset.table(table)
    table = bigquery.Table(table_ref, schema=table_schema(schema))
    client.create_table(table)
    return True


def table_schema(schema):
    # coerce the type of DATA_COL to bytes
    if constants.DATA_COL in schema:
        schema[constants.DATA_COL] = 'BYTES'

    base = [bigquery.SchemaField(k, v) for k, v in schema.items()]
    if constants.DATA_COL not in schema:
        base.append(bigquery.SchemaField(constants.DATA_COL, 'BYTES'))

    return base



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from abc import abstractmethod
from typing import Text, List

from tensorflow_metadata.proto.v0.schema_pb2 import Schema
from tensorflow_metadata.proto.v0.statistics_pb2 import \
    DatasetFeatureStatisticsList

from zenml.core.steps.base_step import BaseStep


class BaseSplitStep(BaseStep):

    def __init__(self,
                 statistics: DatasetFeatureStatisticsList = None,
                 schema: Schema = None):
        """
        Constructor for BaseSplitStep.

        Args:
            statistics: output of a preceding StatisticsGen
            schema: output of a preceding SchemaGen
        """
        self.statistics = statistics
        self.schema = schema

    @abstractmethod
    def partition_fn(self):
        pass

    @abstractmethod
    def get_split_names(self) -> List[Text]:
        pass

    def get_num_splits(self):
        return len(self.get_split_names())
